@@ -74,7 +74,7 @@ bool is_safe_for_add(const Expr &e, int max_depth) {
         } else if (cast->type.bits() == cast->value.type().bits()) {
             return is_safe_for_add(cast->value, max_depth);
         }
-    } else if (Call::as_intrinsic(e, {Call::widening_add, Call::widening_sub})) {
+    } else if (Call::as_intrinsic(e, {Call::widening_add, Call::widening_sub, Call::widen_right_add, Call::widen_right_sub})) {
         return true;
     }
     return false;
@@ -131,20 +131,47 @@ Expr to_rounding_shift(const Call *c) {
         }
         Expr round;
         if (c->is_intrinsic(Call::shift_right)) {
-            round = simplify((make_one(round_type) << max(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2);
+            round = (make_one(round_type) << max(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2;
         } else {
-            round = simplify((make_one(round_type) >> min(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2);
+            round = (make_one(round_type) >> min(cast(b.type().with_bits(round_type.bits()), b), 0)) / 2;
         }
+        // Input expressions are simplified before running find_intrinsics, but b
+        // has been lifted here so we need to lower_intrinsics before simplifying
+        // and re-lifting. Should we move this code into the FindIntrinsics class
+        // to make it easier to lift round?
+        round = lower_intrinsics(round);
+        round = simplify(round);
+        round = find_intrinsics(round);
 
         // We can always handle widening adds.
         if (const Call *add = Call::as_intrinsic(a, {Call::widening_add})) {
-            if (can_prove(lower_intrinsics(add->args[0]) == round)) {
+            if (can_prove(lower_intrinsics(add->args[0] == round))) {
                 return rounding_shift(cast(add->type, add->args[1]), b);
-            } else if (can_prove(lower_intrinsics(add->args[1]) == round)) {
+            } else if (can_prove(lower_intrinsics(add->args[1] == round))) {
                 return rounding_shift(cast(add->type, add->args[0]), b);
             }
         }
 
+        if (const Call *add = Call::as_intrinsic(a, {Call::widen_right_add})) {
+            if (can_prove(lower_intrinsics(add->args[1] == round))) {
+                return rounding_shift(cast(add->type, add->args[0]), b);
+            }
+        }
+        // Also need to handle the annoying case of a reinterpret wrapping a widen_right_add
+        // TODO: this pattern makes me want to change the semantics of this op.
+        if (const Reinterpret *reinterp = a.as<Reinterpret>()) {
+            if (reinterp->type.bits() == reinterp->value.type().bits()) {
+                if (const Call *add = Call::as_intrinsic(reinterp->value, {Call::widen_right_add})) {
+                    if (can_prove(lower_intrinsics(add->args[1] == round))) {
+                        // We expect the first operand to be a reinterpet.
+                        const Reinterpret *reinterp_a = add->args[0].as<Reinterpret>();
+                        internal_assert(reinterp_a) << "Failed: " << add->args[0] << "\n";
+                        return rounding_shift(reinterp_a->value, b);
+                    }
+                }
+            }
+        }
+
         // If it wasn't a widening or saturating add, we might still
         // be able to safely accept the rounding.
         Expr a_less_round = find_and_subtract(a, round);
@@ -199,6 +226,53 @@ protected:
             }
         }
 
+        if (op->type.is_int_or_uint() && op->type.bits() > 8) {
+            // Look for widen_right_add intrinsics.
+            // Yes we do an duplicate code, but we want to check the op->type.code() first,
+            // and the opposite as well.
+            for (halide_type_code_t code : {op->type.code(), halide_type_uint, halide_type_int}) {
+                Type narrow = op->type.narrow().with_code(code);
+                // Pulling casts out of VectorReduce nodes breaks too much codegen, skip for now.
+                Expr narrow_a = (a.node_type() == IRNodeType::VectorReduce) ? Expr() : lossless_cast(narrow, a);
+                Expr narrow_b = (b.node_type() == IRNodeType::VectorReduce) ? Expr() : lossless_cast(narrow, b);
+
+                // This case should have been handled by the above check for widening_add.
+                internal_assert(!(narrow_a.defined() && narrow_b.defined()))
+                    << "find_intrinsics failed to find a widening_add: " << a << " + " << b << "\n";
+
+                if (narrow_a.defined()) {
+                    Expr result;
+                    if (b.type().code() != narrow_a.type().code()) {
+                        // Need to do a safe reinterpret.
+                        Type t = b.type().with_code(code);
+                        result = widen_right_add(reinterpret(t, b), narrow_a);
+                        internal_assert(result.type() != op->type);
+                        result = reinterpret(op->type, result);
+                    } else {
+                        result = widen_right_add(b, narrow_a);
+                    }
+                    internal_assert(result.type() == op->type);
+                    return result;
+                } else if (narrow_b.defined()) {
+                    Expr result;
+                    if (a.type().code() != narrow_b.type().code()) {
+                        // Need to do a safe reinterpret.
+                        Type t = a.type().with_code(code);
+                        result = widen_right_add(reinterpret(t, a), narrow_b);
+                        internal_assert(result.type() != op->type);
+                        result = reinterpret(op->type, result);
+                    } else {
+                        result = widen_right_add(a, narrow_b);
+                    }
+                    internal_assert(result.type() == op->type);
+                    return mutate(result);
+                }
+            }
+        }
+
+        // TODO: there can be widen_right_add + widen_right_add simplification rules.
+        // i.e. widen_right_add(a, b) + widen_right_add(c, d) = (a + c) + widening_add(b, d)
+
         if (a.same_as(op->a) && b.same_as(op->b)) {
             return op;
         } else {
@@ -240,6 +314,32 @@ protected:
             return Add::make(a, negative_b);
         }
 
+        // Run after the lossless_negate check, because we want that to turn into an widen_right_add if relevant.
+        if (op->type.is_int_or_uint() && op->type.bits() > 8) {
+            // Look for widen_right_sub intrinsics.
+            // Yes we do an duplicate code, but we want to check the op->type.code() first,
+            // and the opposite as well.
+            for (halide_type_code_t code : {op->type.code(), halide_type_uint, halide_type_int}) {
+                Type narrow = op->type.narrow().with_code(code);
+                Expr narrow_b = lossless_cast(narrow, b);
+
+                if (narrow_b.defined()) {
+                    Expr result;
+                    if (a.type().code() != narrow_b.type().code()) {
+                        // Need to do a safe reinterpret.
+                        Type t = a.type().with_code(code);
+                        result = widen_right_sub(reinterpret(t, a), narrow_b);
+                        internal_assert(result.type() != op->type);
+                        result = reinterpret(op->type, result);
+                    } else {
+                        result = widen_right_sub(a, narrow_b);
+                    }
+                    internal_assert(result.type() == op->type);
+                    return mutate(result);
+                }
+            }
+        }
+
         if (a.same_as(op->a) && b.same_as(op->b)) {
             return op;
         } else {
@@ -292,6 +392,49 @@ protected:
             return mutate(result);
         }
 
+        if (op->type.is_int_or_uint() && op->type.bits() > 8) {
+            // Look for widen_right_mul intrinsics.
+            // Yes we do an duplicate code, but we want to check the op->type.code() first,
+            // and the opposite as well.
+            for (halide_type_code_t code : {op->type.code(), halide_type_uint, halide_type_int}) {
+                Type narrow = op->type.narrow().with_code(code);
+                Expr narrow_a = lossless_cast(narrow, a);
+                Expr narrow_b = lossless_cast(narrow, b);
+
+                // This case should have been handled by the above check for widening_mul.
+                internal_assert(!(narrow_a.defined() && narrow_b.defined()))
+                    << "find_intrinsics failed to find a widening_mul: " << a << " + " << b << "\n";
+
+                if (narrow_a.defined()) {
+                    Expr result;
+                    if (b.type().code() != narrow_a.type().code()) {
+                        // Need to do a safe reinterpret.
+                        Type t = b.type().with_code(code);
+                        result = widen_right_mul(reinterpret(t, b), narrow_a);
+                        internal_assert(result.type() != op->type);
+                        result = reinterpret(op->type, result);
+                    } else {
+                        result = widen_right_mul(b, narrow_a);
+                    }
+                    internal_assert(result.type() == op->type);
+                    return result;
+                } else if (narrow_b.defined()) {
+                    Expr result;
+                    if (a.type().code() != narrow_b.type().code()) {
+                        // Need to do a safe reinterpret.
+                        Type t = a.type().with_code(code);
+                        result = widen_right_mul(reinterpret(t, a), narrow_b);
+                        internal_assert(result.type() != op->type);
+                        result = reinterpret(op->type, result);
+                    } else {
+                        result = widen_right_mul(a, narrow_b);
+                    }
+                    internal_assert(result.type() == op->type);
+                    return mutate(result);
+                }
+            }
+        }
+
         if (a.same_as(op->a) && b.same_as(op->b)) {
             return op;
         } else {
@@ -410,6 +553,14 @@ protected:
                         saturating_sub(x, y),
                         op->type.is_uint() && is_x_same_uint) ||
 
+                // Saturating narrow patterns.
+                rewrite(max(min(x, upper), lower),
+                        saturating_cast(op->type, x)) ||
+
+                rewrite(min(x, upper),
+                        saturating_cast(op->type, x),
+                        is_uint(x)) ||
+
                 // Averaging patterns
                 //
                 // We have a slight preference for rounding_halving_add over
@@ -447,18 +598,10 @@ protected:
                         rounding_halving_add(x, y),
                         is_x_same_int_or_uint) ||
 
-                rewrite(halving_add(widening_sub(x, y), 1),
-                        rounding_halving_sub(x, y),
-                        is_x_same_int_or_uint) ||
-
                 rewrite(rounding_shift_right(widening_add(x, y), 1),
                         rounding_halving_add(x, y),
                         is_x_same_int_or_uint) ||
 
-                rewrite(rounding_shift_right(widening_sub(x, y), 1),
-                        rounding_halving_sub(x, y),
-                        is_x_same_int_or_uint) ||
-
                 // Multiply-keep-high-bits patterns.
                 rewrite(max(min(shift_right(widening_mul(x, y), z), upper), lower),
                         mul_shift_right(x, y, cast(unsigned_type, z)),
@@ -521,10 +664,6 @@ protected:
                         halving_sub(x, y),
                         is_x_same_int_or_uint) ||
 
-                rewrite(rounding_shift_right(cast(op_type_wide, widening_sub(x, y)), 1),
-                        rounding_halving_sub(x, y),
-                        is_x_same_int_or_uint) ||
-
                 false) {
                 internal_assert(rewrite.result.type() == op->type)
                     << "Rewrite changed type: " << Expr(op) << " -> " << rewrite.result << "\n";
@@ -587,18 +726,105 @@ protected:
             return rewrite.result;
         }
 
+        const int bits = op->type.bits();
+        const int lanes = op->type.lanes();
+        const auto is_x_same_int = op->type.is_int() && is_int(x, bits);
+        const auto is_x_same_uint = op->type.is_uint() && is_uint(x, bits);
+        const auto is_x_same_int_or_uint = is_x_same_int || is_x_same_uint;
+        auto x_y_same_sign = (is_int(x) == is_int(y)) || (is_uint(x) && is_uint(y));
+        Type unsigned_type = op->type.with_code(halide_type_uint);
+        const auto is_x_wider_int_or_uint = (op->type.is_int() && is_int(x, 2 * bits)) || (op->type.is_uint() && is_uint(x, 2 * bits));
+        Type opposite_type = op->type.is_int() ? op->type.with_code(halide_type_uint) : op->type.with_code(halide_type_int);
+        const auto is_x_wider_opposite_int = (op->type.is_int() && is_uint(x, 2 * bits)) || (op->type.is_uint() && is_int(x, 2 * bits));
+
+        if (
+            // Simplify extending patterns.
+            // (x + widen(y)) + widen(z) = x + widening_add(y, z).
+            rewrite(widen_right_add(widen_right_add(x, y), z),
+                    x + widening_add(y, z),
+                    // We only care about integers, this should be trivially true.
+                    is_x_same_int_or_uint) ||
+
+            // (x - widen(y)) - widen(z) = x - widening_add(y, z).
+            rewrite(widen_right_sub(widen_right_sub(x, y), z),
+                    x - widening_add(y, z),
+                    // We only care about integers, this should be trivially true.
+                    is_x_same_int_or_uint) ||
+
+            // (x + widen(y)) - widen(z) = x + cast(t, widening_sub(y, z))
+            // cast (reinterpret) is needed only for uints.
+            rewrite(widen_right_sub(widen_right_add(x, y), z),
+                    x + widening_sub(y, z),
+                    is_x_same_int) ||
+            rewrite(widen_right_sub(widen_right_add(x, y), z),
+                    x + cast(op->type, widening_sub(y, z)),
+                    is_x_same_uint) ||
+
+            // (x - widen(y)) + widen(z) = x + cast(t, widening_sub(z, y))
+            // cast (reinterpret) is needed only for uints.
+            rewrite(widen_right_add(widen_right_sub(x, y), z),
+                    x + widening_sub(z, y),
+                    is_x_same_int) ||
+            rewrite(widen_right_add(widen_right_sub(x, y), z),
+                    x + cast(op->type, widening_sub(z, y)),
+                    is_x_same_uint) ||
+
+            // Saturating patterns.
+            rewrite(saturating_cast(op->type, widening_add(x, y)),
+                    saturating_add(x, y),
+                    is_x_same_int_or_uint) ||
+            rewrite(saturating_cast(op->type, widening_sub(x, y)),
+                    saturating_sub(x, y),
+                    is_x_same_int_or_uint) ||
+            rewrite(saturating_cast(op->type, shift_right(widening_mul(x, y), z)),
+                    mul_shift_right(x, y, cast(unsigned_type, z)),
+                    is_x_same_int_or_uint && x_y_same_sign && is_uint(z)) ||
+            rewrite(saturating_cast(op->type, rounding_shift_right(widening_mul(x, y), z)),
+                    rounding_mul_shift_right(x, y, cast(unsigned_type, z)),
+                    is_x_same_int_or_uint && x_y_same_sign && is_uint(z)) ||
+
+            // Rewrite combinations of deinterleaves into horizontal ops
+            rewrite(widening_add(slice(x, 0, 2, lanes), slice(x, 1, 2, lanes)),
+                    h_add(cast(op->type.with_lanes(lanes * 2), x), lanes)) ||
+            rewrite(widening_add(slice(x, 1, 2, lanes), slice(x, 0, 2, lanes)),
+                    h_add(cast(op->type.with_lanes(lanes * 2), x), lanes)) ||
+
+            // We can remove unnecessary widening if we are then performing a saturating narrow.
+            // This is similar to the logic inside `visit_min_or_max`.
+            (((bits <= 32) &&
+              // Examples:
+              // i8_sat(int16(i8)) -> i8
+              // u8_sat(uint16(u8)) -> u8
+              rewrite(saturating_cast(op->type, cast(op->type.widen(), x)),
+                      x,
+                      is_x_same_int_or_uint)) ||
+             ((bits <= 16) &&
+              // Examples:
+              // i8_sat(int32(i16)) -> i8_sat(i16)
+              // u8_sat(uint32(u16)) -> u8_sat(u16)
+              (rewrite(saturating_cast(op->type, cast(op->type.widen().widen(), x)),
+                       saturating_cast(op->type, x),
+                       is_x_wider_int_or_uint) ||
+               // Examples:
+               // i8_sat(uint32(u16)) -> i8_sat(u16)
+               // u8_sat(int32(i16)) -> i8_sat(i16)
+               rewrite(saturating_cast(op->type, cast(opposite_type.widen().widen(), x)),
+                       saturating_cast(op->type, x),
+                       is_x_wider_opposite_int) ||
+               false))) ||
+            false) {
+            return mutate(rewrite.result);
+        }
+
         if (no_overflow(op->type)) {
             // clang-format off
             if (rewrite(halving_add(x + y, 1), rounding_halving_add(x, y)) ||
                 rewrite(halving_add(x, y + 1), rounding_halving_add(x, y)) ||
                 rewrite(halving_add(x + 1, y), rounding_halving_add(x, y)) ||
-                rewrite(halving_add(x - y, 1), rounding_halving_sub(x, y)) ||
-                rewrite(halving_sub(x + 1, y), rounding_halving_sub(x, y)) ||
                 rewrite(halving_add(x, 1), rounding_shift_right(x, 1)) ||
                 rewrite(shift_right(x + y, 1), halving_add(x, y)) ||
                 rewrite(shift_right(x - y, 1), halving_sub(x, y)) ||
                 rewrite(rounding_shift_right(x + y, 1), rounding_halving_add(x, y)) ||
-                rewrite(rounding_shift_right(x - y, 1), rounding_halving_sub(x, y)) ||
                 false) {
                 return mutate(rewrite.result);
             }
@@ -635,6 +861,7 @@ protected:
                 }
             }
         }
+        // TODO: do we want versions of widen_right_add here?
 
         if (op->is_intrinsic(Call::shift_right) || op->is_intrinsic(Call::shift_left)) {
             // Try to turn this into a widening shift.
@@ -678,6 +905,130 @@ protected:
         }
         return op;
     }
+
+    // Narrow and comparisons between ramps and broadcasts to produce masks that
+    // match the bit-width of the type being selected between or loaded or
+    // stored. We do this late in lowering in this pass instead of in the
+    // simplifier because it messes up the reasoning done by loop partitioning.
+    //
+    // For example if we're selecting between uint8s, and we have the condition:
+    // ramp(x, 1, 16) < broadcast(y)
+    // where x is an Int(32), we can rewrite this to:
+    // ramp(0, 1, 16) < broadcast(y - x)
+    // which can be safely narrowed to:
+    // cast<int8_t>(ramp(0, 1, 16)) < saturating_cast<int8_t>(broadcast(y - x))
+    Expr narrow_predicate(const Expr &p, Type t) {
+        if (t.bits() >= 32) {
+            return p;
+        }
+
+        int lanes = t.lanes();
+
+        if (const Or *op = p.as<Or>()) {
+            return narrow_predicate(op->a, t) || narrow_predicate(op->b, t);
+        } else if (const And *op = p.as<And>()) {
+            return narrow_predicate(op->a, t) && narrow_predicate(op->b, t);
+        } else if (const Not *op = p.as<Not>()) {
+            return !narrow_predicate(op->a, t);
+        }
+
+        const LT *lt = p.as<LT>();
+        const LE *le = p.as<LE>();
+        // Check it's a comparison
+        if (!(le || lt)) {
+            return p;
+        }
+        // Check it's an int32 comparison
+        if ((lt ? lt->a.type() : le->a.type()) != Int(32, lanes)) {
+            return p;
+        }
+
+        auto rewrite = IRMatcher::rewriter(p, Int(32, lanes));
+
+        // Construct predicates which state the ramp can't hit the extreme
+        // values of an int8 or an int16, so that the saturated broadcast has a
+        // value to take on that leaves it clear of the bounds of the ramp. This
+        // is an overconservative condition, but it's hard to imagine cases
+        // where a more precise condition would be necessary.
+        auto min_ramp_lane = min(c0, c0 * (lanes - 1));
+        auto max_ramp_lane = max(c0, c0 * (lanes - 1));
+        auto ramp_fits_in_i8 = min_ramp_lane > -128 && max_ramp_lane < 127;
+        auto ramp_fits_in_i16 = min_ramp_lane > -32768 && max_ramp_lane < 32767;
+        auto saturated_diff_i8 = saturating_cast(Int(8), saturating_sub(x, y));
+        auto saturated_diff_i16 = saturating_cast(Int(16), saturating_sub(x, y));
+
+        if ((t.bits() <= 8 &&
+             // Try to narrow to 8-bit comparisons
+             (rewrite(broadcast(x, lanes) < ramp(y, c0, lanes),
+                      broadcast(saturated_diff_i8, lanes) < cast(Int(8, lanes), ramp(0, c0, lanes)),
+                      ramp_fits_in_i8) ||
+
+              rewrite(ramp(y, c0, lanes) < broadcast(x, lanes),
+                      cast(Int(8, lanes), ramp(0, c0, lanes)) < broadcast(saturated_diff_i8, lanes),
+                      ramp_fits_in_i8) ||
+
+              rewrite(broadcast(x, lanes) <= ramp(y, c0, lanes),
+                      broadcast(saturated_diff_i8, lanes) <= cast(Int(8, lanes), ramp(0, c0, lanes)),
+                      ramp_fits_in_i8) ||
+
+              rewrite(ramp(y, c0, lanes) <= broadcast(x, lanes),
+                      cast(Int(8, lanes), ramp(0, c0, lanes)) <= broadcast(saturated_diff_i8, lanes),
+                      ramp_fits_in_i8))) ||
+
+            // Try to narrow to 16-bit comparisons
+            rewrite(broadcast(x, lanes) < ramp(y, c0, lanes),
+                    broadcast(saturated_diff_i16, lanes) < cast(Int(16, lanes), ramp(0, c0, lanes)),
+                    ramp_fits_in_i16) ||
+
+            rewrite(ramp(y, c0, lanes) < broadcast(x, lanes),
+                    cast(Int(16, lanes), ramp(0, c0, lanes)) < broadcast(saturated_diff_i16, lanes),
+                    ramp_fits_in_i16) ||
+
+            rewrite(broadcast(x, lanes) <= ramp(y, c0, lanes),
+                    broadcast(saturated_diff_i16, lanes) <= cast(Int(16, lanes), ramp(0, c0, lanes)),
+                    ramp_fits_in_i16) ||
+
+            rewrite(ramp(y, c0, lanes) <= broadcast(x, lanes),
+                    cast(Int(16, lanes), ramp(0, c0, lanes)) <= broadcast(saturated_diff_i16, lanes),
+                    ramp_fits_in_i16)) {
+            return rewrite.result;
+        } else {
+            return p;
+        }
+    }
+
+    Expr visit(const Select *op) override {
+        Expr condition = mutate(op->condition);
+        Expr true_value = mutate(op->true_value);
+        Expr false_value = mutate(op->false_value);
+        condition = narrow_predicate(condition, op->type);
+        return Select::make(condition, true_value, false_value);
+    }
+
+    Expr visit(const Load *op) override {
+        Expr predicate = mutate(op->predicate);
+        Expr index = mutate(op->index);
+        predicate = narrow_predicate(predicate, op->type);
+        if (predicate.same_as(op->predicate) && index.same_as(op->index)) {
+            return op;
+        } else {
+            return Load::make(op->type, op->name, std::move(index),
+                              op->image, op->param, std::move(predicate),
+                              op->alignment);
+        }
+    }
+
+    Stmt visit(const Store *op) override {
+        Expr predicate = mutate(op->predicate);
+        Expr value = mutate(op->value);
+        Expr index = mutate(op->index);
+        predicate = narrow_predicate(predicate, value.type());
+        if (predicate.same_as(op->predicate) && value.same_as(op->value) && index.same_as(op->index)) {
+            return op;
+        } else {
+            return Store::make(op->name, std::move(value), std::move(index), op->param, std::move(predicate), op->alignment);
+        }
+    }
 };
 
 // Substitute in let values than have an output vector
@@ -841,6 +1192,18 @@ Expr find_intrinsics(const Expr &e) {
     return expr;
 }
 
+Expr lower_widen_right_add(const Expr &a, const Expr &b) {
+    return a + widen(b);
+}
+
+Expr lower_widen_right_mul(const Expr &a, const Expr &b) {
+    return a * widen(b);
+}
+
+Expr lower_widen_right_sub(const Expr &a, const Expr &b) {
+    return a - widen(b);
+}
+
 Expr lower_widening_add(const Expr &a, const Expr &b) {
     return widen(a) + widen(b);
 }
@@ -888,17 +1251,103 @@ Expr lower_rounding_shift_right(const Expr &a, const Expr &b) {
 }
 
 Expr lower_saturating_add(const Expr &a, const Expr &b) {
-    internal_assert(a.type() == b.type());
     // Lower saturating add without using widening arithmetic, which may require
     // types that aren't supported.
-    return simplify(clamp(a, a.type().min() - min(b, 0), a.type().max() - max(b, 0))) + b;
+    internal_assert(a.type() == b.type());
+    if (a.type().is_float()) {
+        return a + b;
+    } else if (a.type().is_uint()) {
+        Expr sum = a + b;
+        return select(sum < a, a.type().max(), sum);
+    } else if (a.type().is_int()) {
+        Type u = a.type().with_code(halide_type_uint);
+        Expr ua = reinterpret(u, a);
+        Expr ub = reinterpret(u, b);
+        Expr upper = make_const(u, (uint64_t(1) << (a.type().bits() - 1)) - 1);
+        Expr lower = make_const(u, (uint64_t(1) << (a.type().bits() - 1)));
+        Expr sum = ua + ub;
+        // For a 32-bit input, 'sum' is the low 32 bits of the true 33-bit
+        // sum. So it's the true sum, possibly plus 2^32 in the case where the
+        // true sum is supposed to be negative. The true sum is positive when:
+        // a + b >= 0 === a >= -b === a >= ~b + 1 === a > ~b
+        Expr pos_result = min(sum, upper);
+        Expr neg_result = max(sum, lower);
+        return simplify(reinterpret(a.type(), select(~b < a, pos_result, neg_result)));
+    } else {
+        internal_error << "Bad type for saturating_add: " << a.type() << "\n";
+        return Expr();
+    }
 }
 
 Expr lower_saturating_sub(const Expr &a, const Expr &b) {
-    internal_assert(a.type() == b.type());
-    // Lower saturating add without using widening arithmetic, which may require
+    // Lower saturating sub without using widening arithmetic, which may require
     // types that aren't supported.
-    return simplify(clamp(a, a.type().min() + max(b, 0), a.type().max() + min(b, 0))) - b;
+    internal_assert(a.type() == b.type());
+    if (a.type().is_float()) {
+        return a - b;
+    } else if (a.type().is_int()) {
+        // Do the math in unsigned, to avoid overflow in the simplifier.
+        Type u = a.type().with_code(halide_type_uint);
+        Expr ua = reinterpret(u, a);
+        Expr ub = reinterpret(u, b);
+        Expr upper = make_const(u, (uint64_t(1) << (a.type().bits() - 1)) - 1);
+        Expr lower = make_const(u, (uint64_t(1) << (a.type().bits() - 1)));
+        Expr diff = ua - ub;
+        // If a >= b, then diff is the (positive) difference. If a < b then diff
+        // is the (negative) difference plus 2^32 due to wraparound.
+        // We saturate the positive difference to be at most 2^31 - 1
+        Expr pos_diff = min(upper, diff);
+        // and saturate the negative difference to be at least -2^31 + 2^32 = 2^31
+        Expr neg_diff = max(lower, diff);
+        // Then select between them, and cast back to the signed type.
+        return simplify(reinterpret(a.type(), select(b <= a, pos_diff, neg_diff)));
+    } else if (a.type().is_uint()) {
+        return simplify(select(b < a, a - b, make_zero(a.type())));
+    } else {
+        internal_error << "Bad type for saturating_sub: " << a.type() << "\n";
+        return Expr();
+    }
+}
+
+Expr lower_saturating_cast(const Type &t, const Expr &a) {
+    // For float to float, guarantee infinities are always pinned to range.
+    if (t.is_float() && a.type().is_float()) {
+        if (t.bits() < a.type().bits()) {
+            return cast(t, clamp(a, t.min(), t.max()));
+        } else {
+            return clamp(cast(t, a), t.min(), t.max());
+        }
+    } else if (a.type() != t) {
+        // Limits for Int(2^n) or UInt(2^n) are not exactly representable in Float(2^n)
+        if (a.type().is_float() && !t.is_float() && t.bits() >= a.type().bits()) {
+            Expr e = max(a, t.min());  // min values turn out to be always representable
+
+            // This line depends on t.max() rounding upward, which should always
+            // be the case as it is one less than a representable value, thus
+            // the one larger is always the closest.
+            e = select(e >= cast(e.type(), t.max()), t.max(), cast(t, e));
+            return e;
+        } else {
+            Expr min_bound;
+            if (!a.type().is_uint()) {
+                min_bound = lossless_cast(a.type(), t.min());
+            }
+            Expr max_bound = lossless_cast(a.type(), t.max());
+
+            Expr e;
+            if (min_bound.defined() && max_bound.defined()) {
+                e = clamp(a, min_bound, max_bound);
+            } else if (min_bound.defined()) {
+                e = max(a, min_bound);
+            } else if (max_bound.defined()) {
+                e = min(a, max_bound);
+            } else {
+                e = a;
+            }
+            return cast(t, std::move(e));
+        }
+    }
+    return a;
 }
 
 Expr lower_halving_add(const Expr &a, const Expr &b) {
@@ -909,19 +1358,30 @@ Expr lower_halving_add(const Expr &a, const Expr &b) {
 
 Expr lower_halving_sub(const Expr &a, const Expr &b) {
     internal_assert(a.type() == b.type());
-    return (a >> 1) - (b >> 1) - (((b & 1) - (a & 1) + 1) >> 1);
+    Expr e = rounding_halving_add(a, ~b);
+    if (a.type().is_uint()) {
+        // An explanation in 8-bit:
+        //   (x - y) / 2
+        // = (x + 256 - y) / 2 - 128
+        // = (x + (255 - y) + 1) / 2 - 128
+        // = (x + ~y + 1) / 2 - 128
+        // = rounding_halving_add(x, ~y) - 128
+        // = rounding_halving_add(x, ~y) + 128 (due to 2s-complement wrap-around)
+        return e + make_const(e.type(), (uint64_t)1 << (a.type().bits() - 1));
+    } else {
+        // For 2s-complement signed integers, negating is done by flipping the
+        // bits and adding one, so:
+        //   (x - y) / 2
+        // = (x + (-y)) / 2
+        // = (x + (~y + 1)) / 2
+        // = rounding_halving_add(x, ~y)
+        return e;
+    }
 }
 
-// TODO: These should using rounding_shift_right, but lowering that
-// results in double widening and the simplifier doesn't fix it.
 Expr lower_rounding_halving_add(const Expr &a, const Expr &b) {
     internal_assert(a.type() == b.type());
-    return (a >> 1) + (b >> 1) + (((a & 1) + (b & 1) + 1) >> 1);
-}
-
-Expr lower_rounding_halving_sub(const Expr &a, const Expr &b) {
-    internal_assert(a.type() == b.type());
-    return (a >> 1) - (b >> 1) + (((a & 1) - (b & 1) + 1) >> 1);
+    return halving_add(a, b) + ((a ^ b) & 1);
 }
 
 Expr lower_sorted_avg(const Expr &a, const Expr &b) {
@@ -1004,7 +1464,16 @@ Expr lower_rounding_mul_shift_right(const Expr &a, const Expr &b, const Expr &q)
 }
 
 Expr lower_intrinsic(const Call *op) {
-    if (op->is_intrinsic(Call::widening_add)) {
+    if (op->is_intrinsic(Call::widen_right_add)) {
+        internal_assert(op->args.size() == 2);
+        return lower_widen_right_add(op->args[0], op->args[1]);
+    } else if (op->is_intrinsic(Call::widen_right_mul)) {
+        internal_assert(op->args.size() == 2);
+        return lower_widen_right_mul(op->args[0], op->args[1]);
+    } else if (op->is_intrinsic(Call::widen_right_sub)) {
+        internal_assert(op->args.size() == 2);
+        return lower_widen_right_sub(op->args[0], op->args[1]);
+    } else if (op->is_intrinsic(Call::widening_add)) {
         internal_assert(op->args.size() == 2);
         return lower_widening_add(op->args[0], op->args[1]);
     } else if (op->is_intrinsic(Call::widening_mul)) {
@@ -1019,6 +1488,9 @@ Expr lower_intrinsic(const Call *op) {
     } else if (op->is_intrinsic(Call::saturating_sub)) {
         internal_assert(op->args.size() == 2);
         return lower_saturating_sub(op->args[0], op->args[1]);
+    } else if (op->is_intrinsic(Call::saturating_cast)) {
+        internal_assert(op->args.size() == 1);
+        return lower_saturating_cast(op->type, op->args[0]);
     } else if (op->is_intrinsic(Call::widening_shift_left)) {
         internal_assert(op->args.size() == 2);
         return lower_widening_shift_left(op->args[0], op->args[1]);
@@ -1040,9 +1512,6 @@ Expr lower_intrinsic(const Call *op) {
     } else if (op->is_intrinsic(Call::rounding_halving_add)) {
         internal_assert(op->args.size() == 2);
         return lower_rounding_halving_add(op->args[0], op->args[1]);
-    } else if (op->is_intrinsic(Call::rounding_halving_sub)) {
-        internal_assert(op->args.size() == 2);
-        return lower_rounding_halving_sub(op->args[0], op->args[1]);
     } else if (op->is_intrinsic(Call::rounding_mul_shift_right)) {
         internal_assert(op->args.size() == 3);
         return lower_rounding_mul_shift_right(op->args[0], op->args[1], op->args[2]);
