@@ -322,6 +322,15 @@ Stmt remove_promises(const Stmt &s);
  * the tagged expression. If not, returns the expression. */
 Expr unwrap_tags(const Expr &e);
 
+template<typename T>
+struct is_printable_arg {
+    static constexpr bool value = std::is_convertible<T, const char *>::value ||
+                                  std::is_convertible<T, Halide::Expr>::value;
+};
+
+template<typename... Args>
+struct all_are_printable_args : meta_and<is_printable_arg<Args>...> {};
+
 // Secondary args to print can be Exprs or const char *
 inline HALIDE_NO_USER_CODE_INLINE void collect_print_args(std::vector<Expr> &args) {
 }
@@ -342,49 +351,6 @@ Expr requirement_failed_error(Expr condition, const std::vector<Expr> &args);
 
 Expr memoize_tag_helper(Expr result, const std::vector<Expr> &cache_key_values);
 
-/** Compute widen(a) + widen(b). */
-Expr widening_add(Expr a, Expr b);
-/** Compute widen(a) * widen(b). a and b may have different signedness. */
-Expr widening_mul(Expr a, Expr b);
-/** Compute widen(a) - widen(b). The result is always signed. */
-Expr widening_sub(Expr a, Expr b);
-/** Compute widen(a) << b. */
-Expr widening_shift_left(Expr a, Expr b);
-Expr widening_shift_left(Expr a, int b);
-/** Compute widen(a) >> b. */
-Expr widening_shift_right(Expr a, Expr b);
-Expr widening_shift_right(Expr a, int b);
-
-/** Compute saturating_narrow(widening_add(a, (1 >> min(b, 0)) / 2) << b).
- * When b is positive indicating a left shift, the rounding term is zero. */
-Expr rounding_shift_left(Expr a, Expr b);
-Expr rounding_shift_left(Expr a, int b);
-/** Compute saturating_narrow(widening_add(a, (1 << max(b, 0)) / 2) >> b).
- * When b is negative indicating a left shift, the rounding term is zero. */
-Expr rounding_shift_right(Expr a, Expr b);
-Expr rounding_shift_right(Expr a, int b);
-
-/** Compute saturating_narrow(widen(a) + widen(b)) */
-Expr saturating_add(Expr a, Expr b);
-/** Compute saturating_narrow(widen(a) - widen(b)) */
-Expr saturating_sub(Expr a, Expr b);
-
-/** Compute narrow((widen(a) + widen(b)) / 2) */
-Expr halving_add(Expr a, Expr b);
-/** Compute narrow((widen(a) + widen(b) + 1) / 2) */
-Expr rounding_halving_add(Expr a, Expr b);
-/** Compute narrow((widen(a) - widen(b)) / 2) */
-Expr halving_sub(Expr a, Expr b);
-/** Compute narrow((widen(a) - widen(b) + 1) / 2) */
-Expr rounding_halving_sub(Expr a, Expr b);
-
-/** Compute saturating_narrow(shift_right(widening_mul(a, b), q)) */
-Expr mul_shift_right(Expr a, Expr b, Expr q);
-Expr mul_shift_right(Expr a, Expr b, int q);
-/** Compute saturating_narrow(rounding_shift_right(widening_mul(a, b), q)) */
-Expr rounding_mul_shift_right(Expr a, Expr b, Expr q);
-Expr rounding_mul_shift_right(Expr a, Expr b, int q);
-
 }  // namespace Internal
 
 /** Cast an expression to the halide type corresponding to the C++ type T. */
@@ -1022,10 +988,12 @@ Expr floor(Expr x);
 Expr ceil(Expr x);
 
 /** Return the whole number closest to a floating-point expression. If the
- * argument is not floating-point, it is cast to Float(32). The return value
- * is still in floating point, despite being a whole number. On ties, we
- * follow IEEE754 conventions and round to the nearest even number. Vectorizes
- * cleanly. */
+ * argument is not floating-point, it is cast to Float(32). The return value is
+ * still in floating point, despite being a whole number. On ties, we round
+ * towards the nearest even integer. Note that this is not the same as
+ * std::round in C, which rounds away from zero. On platforms without a native
+ * instruction for this, it is emulated, and may be more expensive than
+ * cast<int>(x + 0.5f) or similar. */
 Expr round(Expr x);
 
 /** Return the integer part of a floating-point expression. If the argument is
@@ -1034,21 +1002,21 @@ Expr round(Expr x);
 Expr trunc(Expr x);
 
 /** Returns true if the argument is a Not a Number (NaN). Requires a
-  * floating point argument.  Vectorizes cleanly.
-  * Note that the Expr passed in will be evaluated in strict_float mode,
-  * regardless of whether strict_float mode is enabled in the current Target. */
+ * floating point argument.  Vectorizes cleanly.
+ * Note that the Expr passed in will be evaluated in strict_float mode,
+ * regardless of whether strict_float mode is enabled in the current Target. */
 Expr is_nan(Expr x);
 
 /** Returns true if the argument is Inf or -Inf. Requires a
-  * floating point argument.  Vectorizes cleanly.
-  * Note that the Expr passed in will be evaluated in strict_float mode,
-  * regardless of whether strict_float mode is enabled in the current Target. */
+ * floating point argument.  Vectorizes cleanly.
+ * Note that the Expr passed in will be evaluated in strict_float mode,
+ * regardless of whether strict_float mode is enabled in the current Target. */
 Expr is_inf(Expr x);
 
 /** Returns true if the argument is a finite value (ie, neither NaN nor Inf).
-  * Requires a floating point argument.  Vectorizes cleanly.
-  * Note that the Expr passed in will be evaluated in strict_float mode,
-  * regardless of whether strict_float mode is enabled in the current Target. */
+ * Requires a floating point argument.  Vectorizes cleanly.
+ * Note that the Expr passed in will be evaluated in strict_float mode,
+ * regardless of whether strict_float mode is enabled in the current Target. */
 Expr is_finite(Expr x);
 
 /** Return the fractional part of a floating-point expression. If the argument
@@ -1061,7 +1029,7 @@ Expr reinterpret(Type t, Expr e);
 
 template<typename T>
 Expr reinterpret(Expr e) {
-    return reinterpret(type_of<T>(), e);
+    return reinterpret(type_of<T>(), std::move(e));
 }
 
 /** Return the bitwise and of two expressions (which need not have the
@@ -1563,6 +1531,256 @@ Expr gather(const Expr &e, Args &&...args) {
 }
 // @}
 
+/** Extract a contiguous subsequence of the bits of 'e', starting at the bit
+ * index given by 'lsb', where zero is the least-significant bit, returning a
+ * value of type 't'. Any out-of-range bits requested are filled with zeros.
+ *
+ * extract_bits is especially useful when one wants to load a small vector of a
+ * wide type, and treat it as a larger vector of a smaller type. For example,
+ * loading a vector of 32 uint8 values from a uint32 Func can be done as
+ * follows:
+\code
+f8(x) = extract_bits<uint8_t>(f32(x/4), 8*(x%4));
+f8.align_bounds(x, 4).vectorize(x, 32);
+\endcode
+ * Note that the align_bounds call is critical so that the narrow Exprs are
+ * aligned to the wider Exprs. This makes the x%4 term collapse to a
+ * constant. If f8 is an output Func, then constraining the min value of x to be
+ * a known multiple of four would also be sufficient, e.g. via:
+\code
+f8.output_buffer().dim(0).set_min(0);
+\endcode
+ *
+ * See test/correctness/extract_concat_bits.cpp for a complete example. */
+// @{
+Expr extract_bits(Type t, const Expr &e, const Expr &lsb);
+
+template<typename T>
+Expr extract_bits(const Expr &e, const Expr &lsb) {
+    return extract_bits(type_of<T>(), e, lsb);
+}
+// @}
+
+/** Given a number of Exprs of the same type, concatenate their bits producing a
+ * single Expr of the same type code of the input but with more bits. The
+ * number of arguments must be a power of two.
+ *
+ * concat_bits is especially useful when one wants to treat a Func containing
+ * values of a narrow type as a Func containing fewer values of a wider
+ * type. For example, the following code reinterprets vectors of 32 uint8 values
+ * as a vector of 8 uint32s:
+ *
+\code
+f32(x) = concat_bits({f8(4*x), f8(4*x + 1), f8(4*x + 2), f8(4*x + 3)});
+f32.vectorize(x, 8);
+\endcode
+ *
+ * See test/correctness/extract_concat_bits.cpp for a complete example.
+ */
+Expr concat_bits(const std::vector<Expr> &e);
+
+/** Below is a collection of intrinsics for fixed-point programming. Most of
+ * them can be expressed via other means, but this is more natural for some, as
+ * it avoids ghost widened intermediates that don't (or shouldn't) actually show
+ * up in codegen, and doesn't rely on pattern-matching inside the compiler to
+ * succeed to get good instruction selection.
+ *
+ * The semantics of each call are defined in terms of a non-existent 'widen' and
+ * 'narrow' operators, which stand in for casts that double or halve the
+ * bit-width of a type respectively.
+ */
+
+/** Compute a + widen(b). */
+Expr widen_right_add(Expr a, Expr b);
+
+/** Compute a * widen(b). */
+Expr widen_right_mul(Expr a, Expr b);
+
+/** Compute a - widen(b). */
+Expr widen_right_sub(Expr a, Expr b);
+
+/** Compute widen(a) + widen(b). */
+Expr widening_add(Expr a, Expr b);
+
+/** Compute widen(a) * widen(b). a and b may have different signedness, in which
+ * case the result is signed. */
+Expr widening_mul(Expr a, Expr b);
+
+/** Compute widen(a) - widen(b). The result is always signed. */
+Expr widening_sub(Expr a, Expr b);
+
+/** Compute widen(a) << b. */
+//@{
+Expr widening_shift_left(Expr a, Expr b);
+Expr widening_shift_left(Expr a, int b);
+//@}
+
+/** Compute widen(a) >> b. */
+//@{
+Expr widening_shift_right(Expr a, Expr b);
+Expr widening_shift_right(Expr a, int b);
+//@}
+
+/** Compute saturating_narrow(widening_add(a, (1 >> min(b, 0)) / 2) << b).
+ * When b is positive indicating a left shift, the rounding term is zero. */
+//@{
+Expr rounding_shift_left(Expr a, Expr b);
+Expr rounding_shift_left(Expr a, int b);
+//@}
+
+/** Compute saturating_narrow(widening_add(a, (1 << max(b, 0)) / 2) >> b).
+ * When b is negative indicating a left shift, the rounding term is zero. */
+//@{
+Expr rounding_shift_right(Expr a, Expr b);
+Expr rounding_shift_right(Expr a, int b);
+//@}
+
+/** Compute saturating_narrow(widen(a) + widen(b)) */
+Expr saturating_add(Expr a, Expr b);
+
+/** Compute saturating_narrow(widen(a) - widen(b)) */
+Expr saturating_sub(Expr a, Expr b);
+
+/** Compute narrow((widen(a) + widen(b)) / 2) */
+Expr halving_add(Expr a, Expr b);
+
+/** Compute narrow((widen(a) + widen(b) + 1) / 2) */
+Expr rounding_halving_add(Expr a, Expr b);
+
+/** Compute narrow((widen(a) - widen(b)) / 2) */
+Expr halving_sub(Expr a, Expr b);
+
+/** Compute saturating_narrow(shift_right(widening_mul(a, b), q)) */
+//@{
+Expr mul_shift_right(Expr a, Expr b, Expr q);
+Expr mul_shift_right(Expr a, Expr b, int q);
+//@}
+
+/** Compute saturating_narrow(rounding_shift_right(widening_mul(a, b), q)) */
+//@{
+Expr rounding_mul_shift_right(Expr a, Expr b, Expr q);
+Expr rounding_mul_shift_right(Expr a, Expr b, int q);
+//@}
+
+namespace Internal {
+
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widen_right_add(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widen_right_add(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widen_right_mul(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widen_right_mul(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widen_right_sub(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widen_right_sub(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_add(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widening_add(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_mul(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widening_mul(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_sub(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widening_sub(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_shift_left(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widening_shift_left(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_shift_left(const Expr &a, int b, T * = nullptr) {
+    return Halide::widening_shift_left(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_shift_right(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widening_shift_right(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr widening_shift_right(const Expr &a, int b, T * = nullptr) {
+    return Halide::widening_shift_right(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_shift_left(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::widening_shift_left(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_shift_left(const Expr &a, int b, T * = nullptr) {
+    return Halide::widening_shift_left(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_shift_right(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::rounding_shift_right(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_shift_right(const Expr &a, int b, T * = nullptr) {
+    return Halide::rounding_shift_right(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr saturating_add(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::saturating_add(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr saturating_sub(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::saturating_sub(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr halving_add(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::halving_add(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_halving_add(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::rounding_halving_add(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr halving_sub(const Expr &a, const Expr &b, T * = nullptr) {
+    return Halide::halving_sub(a, b);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr mul_shift_right(const Expr &a, const Expr &b, const Expr &q, T * = nullptr) {
+    return Halide::mul_shift_right(a, b, q);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr mul_shift_right(const Expr &a, const Expr &b, int q, T * = nullptr) {
+    return Halide::mul_shift_right(a, b, q);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_mul_shift_right(const Expr &a, const Expr &b, const Expr &q, T * = nullptr) {
+    return Halide::rounding_mul_shift_right(a, b, q);
+}
+template<typename T = void>
+HALIDE_ATTRIBUTE_DEPRECATED("This function has been moved out of the Halide::Internal:: namespace into Halide::")
+Expr rounding_mul_shift_right(const Expr &a, const Expr &b, int q, T * = nullptr) {
+    return Halide::rounding_mul_shift_right(a, b, q);
+}
+}  // namespace Internal
+
 }  // namespace Halide
 
 #endif
