@@ -378,8 +378,18 @@ void emit_file(const llvm::Module &module_in, Internal::LLVMOStream &out,
     // Make sure things marked as always-inline get inlined
     pass_manager.add(llvm::createAlwaysInlinerLegacyPass());
 
+#if LLVM_VERSION < 170
     // Remove any stale debug info
+    //
+    // Note: this pass was added in https://github.com/halide/Halide/pull/2060;
+    // based on the comments, it looks like it was an attempt to fix an error,
+    // but didn't actually fix it, and (apparently) just got left in?
+    //
+    // There is a 'new' equivalent that we could add in the optimization pass
+    // in Codegen_LLVM.cpp, but since this seems to be have added in error,
+    // we're just going to elide it for LLVM >= 17.0
     pass_manager.add(llvm::createStripDeadDebugInfoPass());
+#endif
 
     // Enable symbol rewriting. This allows code outside libHalide to
     // use symbol rewriting when compiling Halide code (for example, by
