@@ -62,6 +62,7 @@
 #include "SkipStages.h"
 #include "SlidingWindow.h"
 #include "SplitTuples.h"
+#include "StageStridedLoads.h"
 #include "StorageFlattening.h"
 #include "StorageFolding.h"
 #include "StrictifyFloat.h"
@@ -252,6 +253,7 @@ void lower_impl(const vector<Function> &output_funcs,
     // OpenGL relies on GPU var canonicalization occurring before
     // storage flattening.
     if (t.has_gpu_feature() ||
+        t.has_feature(Target::Vulkan) ||
         t.has_feature(Target::OpenGLCompute)) {
         debug(1) << "Canonicalizing GPU var names...\n";
         s = canonicalize_gpu_vars(s);
@@ -320,6 +322,7 @@ void lower_impl(const vector<Function> &output_funcs,
     log("Lowering after vectorizing:", s);
 
     if (t.has_gpu_feature() ||
+        t.has_feature(Target::Vulkan) ||
         t.has_feature(Target::OpenGLCompute)) {
         debug(1) << "Injecting per-block gpu synchronization...\n";
         s = fuse_gpu_thread_loops(s);
@@ -336,6 +339,10 @@ void lower_impl(const vector<Function> &output_funcs,
     s = simplify(s);
     log("Lowering after partitioning loops:", s);
 
+    debug(1) << "Staging strided loads...\n";
+    s = stage_strided_loads(s);
+    log("Lowering after staging strided loads:", s);
+
     debug(1) << "Trimming loops to the region over which they do something...\n";
     s = trim_no_ops(s);
     log("Lowering after loop trimming:", s);
@@ -571,7 +578,7 @@ Module lower(const vector<Function> &output_funcs,
              const vector<Stmt> &requirements,
              bool trace_pipeline,
              const vector<IRMutator *> &custom_passes) {
-    Module result_module{extract_namespaces(pipeline_name), t};
+    Module result_module{strip_namespaces(pipeline_name), t};
     run_with_large_stack([&]() {
         lower_impl(output_funcs, pipeline_name, t, args, linkage_type, requirements, trace_pipeline, custom_passes, result_module);
     });
