@@ -13,7 +13,6 @@
 
 #include "Argument.h"
 #include "Expr.h"
-#include "ExternalCode.h"
 #include "Function.h"  // for NameMangling
 #include "ModulusRemainder.h"
 
@@ -32,6 +31,7 @@ enum class OutputFileType {
     compiler_log,
     cpp_stub,
     featurization,
+    function_info_header,
     llvm_assembly,
     object,
     python_extension,
@@ -43,42 +43,6 @@ enum class OutputFileType {
     stmt_html,
 };
 
-class HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output") Output {
-public:
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType assembly = OutputFileType::assembly;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType bitcode = OutputFileType::bitcode;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType c_header = OutputFileType::c_header;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType c_source = OutputFileType::c_source;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType compiler_log = OutputFileType::compiler_log;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType cpp_stub = OutputFileType::cpp_stub;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType featurization = OutputFileType::featurization;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType llvm_assembly = OutputFileType::llvm_assembly;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType object = OutputFileType::object;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType python_extension = OutputFileType::python_extension;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType pytorch_wrapper = OutputFileType::pytorch_wrapper;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType registration = OutputFileType::registration;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType schedule = OutputFileType::schedule;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType static_library = OutputFileType::static_library;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType stmt = OutputFileType::stmt;
-    HALIDE_ATTRIBUTE_DEPRECATED("Use OutputFileType instead of Output")
-    static constexpr OutputFileType stmt_html = OutputFileType::stmt_html;
-};  // namespace Output
-
 /** Type of linkage a function in a lowered Halide module can have.
     Also controls whether auxiliary functions and metadata are generated. */
 enum class LinkageType {
@@ -167,13 +131,15 @@ class CompilerLogger;
 
 struct AutoSchedulerResults;
 
+using MetadataNameMap = std::map<std::string, std::string>;
+
 /** A halide module. This represents IR containing lowered function
  * definitions and buffers. */
 class Module {
     Internal::IntrusivePtr<Internal::ModuleContents> contents;
 
 public:
-    Module(const std::string &name, const Target &target);
+    Module(const std::string &name, const Target &target, const MetadataNameMap &metadata_name_map = {});
 
     /** Get the target this module has been lowered for. */
     const Target &target() const;
@@ -195,7 +161,6 @@ public:
     const std::vector<Internal::LoweredFunc> &functions() const;
     std::vector<Internal::LoweredFunc> &functions();
     const std::vector<Module> &submodules() const;
-    const std::vector<ExternalCode> &external_code() const;
     // @}
 
     /** Return the function with the given name. If no such function
@@ -207,7 +172,6 @@ public:
     void append(const Buffer<void> &buffer);
     void append(const Internal::LoweredFunc &function);
     void append(const Module &module);
-    void append(const ExternalCode &external_code);
     // @}
 
     /** Compile a halide Module to variety of outputs, depending on
@@ -228,7 +192,7 @@ public:
     void remap_metadata_name(const std::string &from, const std::string &to) const;
 
     /** Retrieve the metadata name map. */
-    std::map<std::string, std::string> get_metadata_name_map() const;
+    MetadataNameMap get_metadata_name_map() const;
 
     /** Set the AutoSchedulerResults for the Module. It is an error to call this
      * multiple times for a given Module. */
