@@ -105,10 +105,10 @@ ConstantInterval add(const ConstantInterval &a, const ConstantInterval &b) {
     result.min_defined = a.has_lower_bound() && b.has_lower_bound();
     result.max_defined = a.has_upper_bound() && b.has_upper_bound();
     if (result.has_lower_bound()) {
-        result.min = a.min + b.min;
+        result.min_defined = add_with_overflow(64, a.min, b.min, &result.min);
     }
     if (result.has_upper_bound()) {
-        result.max = a.max + b.max;
+        result.max_defined = add_with_overflow(64, a.max, b.max, &result.max);
     }
     return result;
 }
@@ -120,14 +120,27 @@ ConstantInterval add(const ConstantInterval &a, int64_t b) {
 ConstantInterval negate(const ConstantInterval &r) {
     ConstantInterval result;
     result.min_defined = r.has_upper_bound();
-    result.min = r.has_upper_bound() ? -r.max : 0;
+    if (result.min_defined) {
+        result.min_defined = sub_with_overflow(64, 0, r.max, &result.min);
+    }
     result.max_defined = r.has_lower_bound();
-    result.max = r.has_lower_bound() ? -r.min : 0;
+    if (result.max_defined) {
+        result.max_defined = sub_with_overflow(64, 0, r.min, &result.max);
+    }
     return result;
 }
 
 ConstantInterval sub(const ConstantInterval &a, const ConstantInterval &b) {
-    return add(a, negate(b));
+    ConstantInterval result;
+    result.min_defined = a.has_lower_bound() && b.has_lower_bound();
+    result.max_defined = a.has_upper_bound() && b.has_upper_bound();
+    if (result.has_lower_bound()) {
+        result.min_defined = sub_with_overflow(64, a.min, b.max, &result.min);
+    }
+    if (result.has_upper_bound()) {
+        result.max_defined = sub_with_overflow(64, a.max, b.min, &result.max);
+    }
+    return result;
 }
 
 ConstantInterval sub(const ConstantInterval &a, int64_t b) {
@@ -160,19 +173,20 @@ ConstantInterval multiply(const ConstantInterval &a, const ConstantInterval &b)
     int64_t bounds[4];
     int64_t *bounds_begin = &bounds[0];
     int64_t *bounds_end = &bounds[0];
+    bool no_overflow = true;
     if (a.has_lower_bound() && b.has_lower_bound()) {
-        *bounds_end++ = a.min * b.min;
+        no_overflow = no_overflow && mul_with_overflow(64, a.min, b.min, bounds_end++);
     }
     if (a.has_lower_bound() && b.has_upper_bound()) {
-        *bounds_end++ = a.min * b.max;
+        no_overflow = no_overflow && mul_with_overflow(64, a.min, b.max, bounds_end++);
     }
     if (a.has_upper_bound() && b.has_lower_bound()) {
-        *bounds_end++ = a.max * b.min;
+        no_overflow = no_overflow && mul_with_overflow(64, a.max, b.min, bounds_end++);
     }
     if (a.has_upper_bound() && b.has_upper_bound()) {
-        *bounds_end++ = a.max * b.max;
+        no_overflow = no_overflow && mul_with_overflow(64, a.max, b.max, bounds_end++);
     }
-    if (bounds_begin != bounds_end) {
+    if (no_overflow && (bounds_begin != bounds_end)) {
         ConstantInterval result = {
             *std::min_element(bounds_begin, bounds_end),
             *std::max_element(bounds_begin, bounds_end),
@@ -211,7 +225,7 @@ ConstantInterval divide(const ConstantInterval &a, int64_t b) {
         result.min = div_imp(result.min, b);
     }
     if (result.has_upper_bound()) {
-        result.max = div_imp(result.max + b - 1, b);
+        result.max = div_imp(result.max - 1, b) + 1;
     }
     return result;
 }
@@ -259,6 +273,10 @@ class DerivativeBounds : public IRVisitor {
         }
     }
 
+    void visit(const Reinterpret *op) override {
+        result = ConstantInterval::everything();
+    }
+
     void visit(const Variable *op) override {
         if (op->name == var) {
             result = ConstantInterval::single_point(1);
@@ -476,7 +494,8 @@ class DerivativeBounds : public IRVisitor {
         }
 
         if (op->is_intrinsic(Call::unsafe_promise_clamped) ||
-            op->is_intrinsic(Call::promise_clamped)) {
+            op->is_intrinsic(Call::promise_clamped) ||
+            op->is_intrinsic(Call::saturating_cast)) {
             op->args[0].accept(this);
             return;
         }
