@@ -80,6 +80,11 @@ class ExprCost : public IRVisitor {
         arith += 1;
     }
 
+    void visit(const Reinterpret *op) override {
+        op->value.accept(this);
+        // `Reinterpret` is a no-op and does *not* incur any cost.
+    }
+
     template<typename T>
     void visit_binary_operator(const T *op, int op_cost) {
         op->a.accept(this);
@@ -219,17 +224,18 @@ class ExprCost : public IRVisitor {
             // TODO: Improve the cost model. In some architectures (e.g. ARM or
             // NEON), count_leading_zeros should be as cheap as bitwise ops.
             // div_round_to_zero and mod_round_to_zero can also get fairly expensive.
-            if (call->is_intrinsic(Call::reinterpret) || call->is_intrinsic(Call::bitwise_and) ||
+            if (call->is_intrinsic(Call::bitwise_and) ||
                 call->is_intrinsic(Call::bitwise_not) || call->is_intrinsic(Call::bitwise_xor) ||
                 call->is_intrinsic(Call::bitwise_or) || call->is_intrinsic(Call::shift_left) ||
                 call->is_intrinsic(Call::shift_right) || call->is_intrinsic(Call::div_round_to_zero) ||
                 call->is_intrinsic(Call::mod_round_to_zero) || call->is_intrinsic(Call::undef) ||
-                call->is_intrinsic(Call::mux)) {
+                call->is_intrinsic(Call::mux) || call->is_intrinsic(Call::round)) {
                 arith += 1;
             } else if (call->is_intrinsic(Call::abs) || call->is_intrinsic(Call::absd) ||
                        call->is_intrinsic(Call::lerp) || call->is_intrinsic(Call::random) ||
                        call->is_intrinsic(Call::count_leading_zeros) ||
-                       call->is_intrinsic(Call::count_trailing_zeros)) {
+                       call->is_intrinsic(Call::count_trailing_zeros) ||
+                       call->is_intrinsic(Call::saturating_cast)) {
                 arith += 5;
             } else if (Call::as_tag(call)) {
                 // Tags do not result in actual operations.
@@ -371,7 +377,7 @@ Expr get_func_value_size(const Function &f) {
 
 Cost compute_expr_cost(Expr expr) {
     // TODO: Handle likely
-    //expr = LikelyExpression().mutate(expr);
+    // expr = LikelyExpression().mutate(expr);
     expr = simplify(expr);
     ExprCost cost_visitor;
     expr.accept(&cost_visitor);
@@ -380,7 +386,7 @@ Cost compute_expr_cost(Expr expr) {
 
 map<string, Expr> compute_expr_detailed_byte_loads(Expr expr) {
     // TODO: Handle likely
-    //expr = LikelyExpression().mutate(expr);
+    // expr = LikelyExpression().mutate(expr);
     expr = simplify(expr);
     ExprCost cost_visitor;
     expr.accept(&cost_visitor);
@@ -508,7 +514,7 @@ RegionCosts::stage_detailed_load_costs(const string &func, int stage,
 
     if (curr_f.has_extern_definition()) {
         // TODO(psuriana): We need a better cost for extern function
-        //load_costs.emplace(func, Int(64).max());
+        // load_costs.emplace(func, Int(64).max());
         load_costs.emplace(func, Expr());
     } else {
         Definition def = get_stage_definition(curr_f, stage);
