@@ -313,7 +313,7 @@ std::string extract_namespaces(const std::string &name, std::vector<std::string>
     return result;
 }
 
-std::string extract_namespaces(const std::string &name) {
+std::string strip_namespaces(const std::string &name) {
     std::vector<std::string> unused;
     return extract_namespaces(name, unused);
 }
@@ -522,6 +522,27 @@ bool add_would_overflow(int bits, int64_t a, int64_t b) {
             (b < 0 && a < min_val - b));   // (a + b) < min_val, rewritten to avoid overflow
 }
 
+bool add_with_overflow(int bits, int64_t a, int64_t b, int64_t *result) {
+#ifndef _MSC_VER
+    if (bits == 64) {
+        static_assert(sizeof(long long) == sizeof(int64_t));
+        bool flag = __builtin_saddll_overflow(a, b, (long long *)result);
+        if (flag) {
+            // Overflowed 64 bits
+            *result = 0;
+        }
+        return !flag;
+    }
+#endif
+    if (add_would_overflow(bits, a, b)) {
+        *result = 0;
+        return false;
+    } else {
+        *result = a + b;
+        return true;
+    }
+}
+
 bool sub_would_overflow(int bits, int64_t a, int64_t b) {
     int64_t max_val = 0x7fffffffffffffffLL >> (64 - bits);
     int64_t min_val = -max_val - 1;
@@ -529,6 +550,27 @@ bool sub_would_overflow(int bits, int64_t a, int64_t b) {
             (b > 0 && a < min_val + b));   // (a - b) < min_val, rewritten to avoid overflow
 }
 
+bool sub_with_overflow(int bits, int64_t a, int64_t b, int64_t *result) {
+#ifndef _MSC_VER
+    if (bits == 64) {
+        static_assert(sizeof(long long) == sizeof(int64_t));
+        bool flag = __builtin_ssubll_overflow(a, b, (long long *)result);
+        if (flag) {
+            // Overflowed 64 bits
+            *result = 0;
+        }
+        return !flag;
+    }
+#endif
+    if (sub_would_overflow(bits, a, b)) {
+        *result = 0;
+        return false;
+    } else {
+        *result = a - b;
+        return true;
+    }
+}
+
 bool mul_would_overflow(int bits, int64_t a, int64_t b) {
     int64_t max_val = 0x7fffffffffffffffLL >> (64 - bits);
     int64_t min_val = -max_val - 1;
@@ -548,6 +590,27 @@ bool mul_would_overflow(int bits, int64_t a, int64_t b) {
     }
 }
 
+bool mul_with_overflow(int bits, int64_t a, int64_t b, int64_t *result) {
+#ifndef _MSC_VER
+    if (bits == 64) {
+        static_assert(sizeof(long long) == sizeof(int64_t));
+        bool flag = __builtin_smulll_overflow(a, b, (long long *)result);
+        if (flag) {
+            // Overflowed 64 bits
+            *result = 0;
+        }
+        return !flag;
+    }
+#endif
+    if (mul_would_overflow(bits, a, b)) {
+        *result = 0;
+        return false;
+    } else {
+        *result = a * b;
+        return true;
+    }
+}
+
 struct TickStackEntry {
     std::chrono::time_point<std::chrono::high_resolution_clock> time;
     string file;
@@ -661,6 +724,19 @@ size_t get_compiler_stack_size() {
 
 namespace Internal {
 
+#if defined(HALIDE_INTERNAL_USING_ASAN) || defined(__ANDROID__)
+// If we are compiling under ASAN,  we will get a zillion warnings about
+// ASAN not supporting makecontext/swapcontext and the possibility of
+// false positives.
+//
+// If we are building for Android, well, it apparently doesn't provide
+// makecontext() / swapcontext(), despite being posixy
+#define MAKECONTEXT_OK 0
+#else
+#define MAKECONTEXT_OK 1
+#endif
+
+#if MAKECONTEXT_OK
 namespace {
 // We can't reliably pass arguments through makecontext, because
 // the calling convention involves an invalid function pointer
@@ -668,6 +744,7 @@ namespace {
 // platforms, so we use a thread local to pass arguments.
 thread_local void *run_with_large_stack_arg = nullptr;
 }  // namespace
+#endif
 
 void run_with_large_stack(const std::function<void()> &action) {
     if (stack_size.size == 0) {
@@ -677,7 +754,6 @@ void run_with_large_stack(const std::function<void()> &action) {
     }
 
 #if _WIN32
-
     // Only exists for its address, which is used to compute remaining stack space.
     ULONG_PTR approx_stack_pos;
 
@@ -719,6 +795,11 @@ void run_with_large_stack(const std::function<void()> &action) {
 #else
     // On posixy systems we have makecontext / swapcontext
 
+#if !MAKECONTEXT_OK
+    action();
+    return;
+#else
+
 #ifdef HALIDE_WITH_EXCEPTIONS
     struct Args {
         const std::function<void()> &run;
@@ -783,6 +864,8 @@ void run_with_large_stack(const std::function<void()> &action) {
     }
 #endif
 
+#endif  // not ADDRESS_SANITIZER
+
 #endif
 }
 
