@@ -307,42 +307,44 @@ public:
 
 }  // namespace
 
-void LoadJacobian::dump(const char *prefix) const {
+void LoadJacobian::dump(std::ostream &os, const char *prefix) const {
     if (count() > 1) {
-        aslog(0) << prefix << count() << " x\n";
+        os << prefix << count() << " x\n";
     }
     for (size_t i = 0; i < producer_storage_dims(); i++) {
-        aslog(0) << prefix << "  [";
+        os << prefix << "  [";
 
         for (size_t j = 0; j < consumer_loop_dims(); j++) {
             const auto &c = (*this)(i, j);
             if (!c.exists) {
-                aslog(0) << " _  ";
+                os << " _  ";
             } else if (c.denominator == 1) {
-                aslog(0) << " " << c.numerator << "  ";
+                os << " " << c.numerator << "  ";
             } else {
-                aslog(0) << c.numerator << "/" << c.denominator << " ";
+                os << c.numerator << "/" << c.denominator << " ";
             }
         }
-        aslog(0) << "]\n";
+        os << "]\n";
     }
-    aslog(0) << "\n";
+    os << "\n";
 }
 
 void BoundContents::validate() const {
     for (int i = 0; i < layout->total_size; i++) {
         auto p = data()[i];
         if (p.max() < p.min()) {
-            aslog(0) << "Bad bounds object:\n";
+            std::ostringstream err;
+            err << "Bad bounds object:\n";
             for (int j = 0; j < layout->total_size; j++) {
                 if (i == j) {
-                    aslog(0) << "=> ";
+                    err << "=> ";
                 } else {
-                    aslog(0) << "   ";
+                    err << "   ";
                 }
-                aslog(0) << j << ": " << data()[j].min() << ", " << data()[j].max() << "\n";
+                err << j << ": " << data()[j].min() << ", " << data()[j].max() << "\n";
             }
-            internal_error << "Aborting";
+            err << "Aborting";
+            internal_error << err.str();
         }
     }
 }
@@ -570,7 +572,7 @@ bool depends_on_estimate(const Expr &expr) {
     return dependency_checker.found_estimate;
 }
 
-FunctionDAG::FunctionDAG(const vector<Function> &outputs, const MachineParams &params, const Target &target) {
+FunctionDAG::FunctionDAG(const vector<Function> &outputs, const Target &target) {
     map<string, Function> env = build_environment(outputs);
 
     // A mutator to apply parameter estimates to the expressions
@@ -816,6 +818,11 @@ FunctionDAG::FunctionDAG(const vector<Function> &outputs, const MachineParams &p
                     check_type(op->type);
                 }
 
+                void visit(const Reinterpret *op) override {
+                    IRVisitor::visit(op);
+                    check_type(op->type);
+                }
+
                 void check_type(Type t) {
                     if (t.bits() > 1 &&
                         (!narrowest_type.bits() ||
@@ -867,8 +874,11 @@ FunctionDAG::FunctionDAG(const vector<Function> &outputs, const MachineParams &p
                 // Get the bounds estimate
                 map<string, Span> estimates;
                 for (const auto &b : consumer.schedule().estimates()) {
-                    int64_t i_min = *as_const_int(b.min);
-                    int64_t i_extent = *as_const_int(b.extent);
+                    const int64_t *i_min = as_const_int(b.min);
+                    const int64_t *i_extent = as_const_int(b.extent);
+                    user_assert(i_min && i_extent)
+                        << "Min/extent of estimate or bound is not constant in \"" << consumer.name()
+                        << "\", var:" << b.var << ", min:" << b.min << ", extent:" << b.extent;
 
                     if ((false)) {  // Intentional dead code. Extra parens to pacify clang-tidy.
                         // Some methods we compare to compile for
@@ -884,9 +894,9 @@ FunctionDAG::FunctionDAG(const vector<Function> &outputs, const MachineParams &p
                         // like unroll across color channels, so
                         // it affects the scheduling space.
                         Func(node.func).bound(b.var, b.min, b.extent);
-                        estimates[b.var] = Span(i_min, i_min + i_extent - 1, true);
+                        estimates[b.var] = Span(*i_min, *i_min + *i_extent - 1, true);
                     } else {
-                        estimates[b.var] = Span(i_min, i_min + i_extent - 1, false);
+                        estimates[b.var] = Span(*i_min, *i_min + *i_extent - 1, false);
                     }
                 }
                 for (const auto &b : consumer.schedule().bounds()) {
@@ -952,7 +962,7 @@ FunctionDAG::FunctionDAG(const vector<Function> &outputs, const MachineParams &p
             }
 
             node.is_wrapper = node.func.is_wrapper();
-            node.is_input = !node.func.has_update_definition() && node.is_wrapper && !any_incoming_edges;
+            node.is_input = !node.is_output && !node.func.has_update_definition() && node.is_wrapper && !any_incoming_edges;
             node.dimensions = node.func.dimensions();
         }
     }
@@ -1039,8 +1049,7 @@ void FunctionDAG::featurize() {
     }
 }
 
-template<typename OS>
-void FunctionDAG::dump_internal(OS &os) const {
+void FunctionDAG::dump(std::ostream &os) const {
     for (const Node &n : nodes) {
         os << "Node: " << n.func.name() << "\n"
            << "  Symbolic region required: \n";
@@ -1076,21 +1085,11 @@ void FunctionDAG::dump_internal(OS &os) const {
 
         os << "  Load Jacobians:\n";
         for (const auto &jac : e.load_jacobians) {
-            jac.dump("  ");
+            jac.dump(os, "  ");
         }
     }
 }
 
-void FunctionDAG::dump() const {
-    auto os = aslog(0);
-    dump_internal(os);
-}
-
-std::ostream &FunctionDAG::dump(std::ostream &os) const {
-    dump_internal(os);
-    return os;
-}
-
 }  // namespace Autoscheduler
 }  // namespace Internal
 }  // namespace Halide
