@@ -27,6 +27,8 @@ using std::string;
 using std::unique_ptr;
 using std::vector;
 
+struct Adams2019Params;
+
 // First we have various utility classes.
 
 // An optional rational type used when analyzing memory dependencies.
@@ -205,7 +207,7 @@ public:
         return result;
     }
 
-    void dump(const char *prefix) const;
+    void dump(std::ostream &os, const char *prefix) const;
 };
 
 // Classes to represent a concrete set of bounds for a Func. A Span is
@@ -563,18 +565,14 @@ struct FunctionDAG {
 
     // Create the function DAG, and do all the dependency and cost
     // analysis. This is done once up-front before the tree search.
-    FunctionDAG(const vector<Function> &outputs, const MachineParams &params, const Target &target);
+    FunctionDAG(const vector<Function> &outputs, const Target &target);
 
-    void dump() const;
-    std::ostream &dump(std::ostream &os) const;
+    void dump(std::ostream &os) const;
 
 private:
     // Compute the featurization for the entire DAG
     void featurize();
 
-    template<typename OS>
-    void dump_internal(OS &os) const;
-
 public:
     // This class uses a lot of internal pointers, so we'll make it uncopyable/unmovable.
     FunctionDAG(const FunctionDAG &other) = delete;
