@@ -3,7 +3,7 @@
 #include "device_interface.h"
 #include "gpu_context_common.h"
 #include "printer.h"
-#include "scoped_spin_lock.h"
+#include "scoped_mutex_lock.h"
 
 #include "objc_support.h"
 
@@ -280,7 +280,7 @@ WEAK mtl_device *get_default_mtl_device() {
 
 extern WEAK halide_device_interface_t metal_device_interface;
 
-volatile ScopedSpinLock::AtomicFlag WEAK thread_lock = 0;
+WEAK halide_mutex thread_lock;
 WEAK mtl_device *device;
 WEAK mtl_command_queue *queue;
 
@@ -297,8 +297,8 @@ WEAK bool metal_api_supports_set_bytes;
 WEAK mtl_device *metal_api_checked_device;
 
 namespace {
-int do_device_to_device_copy(void *user_context, mtl_blit_command_encoder *encoder,
-                             const device_copy &c, uint64_t src_offset, uint64_t dst_offset, int d) {
+void do_device_to_device_copy(void *user_context, mtl_blit_command_encoder *encoder,
+                              const device_copy &c, uint64_t src_offset, uint64_t dst_offset, int d) {
     if (d == 0) {
         buffer_to_buffer_1d_copy(encoder, ((device_handle *)c.src)->buf, c.src_begin + src_offset,
                                  ((device_handle *)c.dst)->buf, dst_offset, c.chunk_size);
@@ -307,15 +307,11 @@ int do_device_to_device_copy(void *user_context, mtl_blit_command_encoder *encod
         // device_buffer_utils.h does not do so either.
         uint64_t src_off = 0, dst_off = 0;
         for (uint64_t i = 0; i < c.extent[d - 1]; i++) {
-            int err = do_device_to_device_copy(user_context, encoder, c, src_offset + src_off, dst_offset + dst_off, d - 1);
+            do_device_to_device_copy(user_context, encoder, c, src_offset + src_off, dst_offset + dst_off, d - 1);
             dst_off += c.dst_stride_bytes[d - 1];
             src_off += c.src_stride_bytes[d - 1];
-            if (err) {
-                return err;
-            }
         }
     }
-    return 0;
 }
 }  // namespace
 
@@ -330,7 +326,7 @@ using namespace Halide::Runtime::Internal::Metal;
 extern "C" {
 
 // The default implementation of halide_metal_acquire_context uses the global
-// pointers above, and serializes access with a spin lock.
+// pointers above, and serializes access with a mutex.
 // Overriding implementations of acquire/release must implement the following
 // behavior:
 // - halide_acquire_metal_context should always store a valid device/command
@@ -341,8 +337,7 @@ extern "C" {
 WEAK int halide_metal_acquire_context(void *user_context, mtl_device **device_ret,
                                       mtl_command_queue **queue_ret, bool create) {
     halide_debug_assert(user_context, &thread_lock != nullptr);
-    while (__atomic_test_and_set(&thread_lock, __ATOMIC_ACQUIRE)) {
-    }
+    halide_mutex_lock(&thread_lock);
 
 #ifdef DEBUG_RUNTIME
     halide_start_clock(user_context);
@@ -352,36 +347,36 @@ WEAK int halide_metal_acquire_context(void *user_context, mtl_device **device_re
         debug(user_context) << "Metal - Allocating: MTLCreateSystemDefaultDevice\n";
         device = get_default_mtl_device();
         if (device == nullptr) {
-            error(user_context) << "Metal: cannot allocate system default device.\n";
-            __atomic_clear(&thread_lock, __ATOMIC_RELEASE);
-            return -1;
+            halide_mutex_unlock(&thread_lock);
+            error(user_context) << "halide_metal_acquire_context: cannot allocate system default device.";
+            return halide_error_code_generic_error;
         }
         debug(user_context) << "Metal - Allocating: new_command_queue\n";
         queue = new_command_queue(device);
         if (queue == nullptr) {
-            error(user_context) << "Metal: cannot allocate command queue.\n";
             release_ns_object(device);
             device = nullptr;
-            __atomic_clear(&thread_lock, __ATOMIC_RELEASE);
-            return -1;
+            halide_mutex_unlock(&thread_lock);
+            error(user_context) << "halide_metal_acquire_context: cannot allocate command queue.";
+            return halide_error_code_generic_error;
         }
     }
 
     // If the device has already been initialized,
     // ensure the queue has as well.
     if (device != nullptr && queue == nullptr) {
-        error(user_context) << "halide_metal_acquire_context: device initialized but queue is not.\n";
-        return -1;
+        error(user_context) << "halide_metal_acquire_context: device initialized but queue is not.";
+        return halide_error_code_generic_error;
     }
 
     *device_ret = device;
     *queue_ret = queue;
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_release_context(void *user_context) {
-    __atomic_clear(&thread_lock, __ATOMIC_RELEASE);
-    return 0;
+    halide_mutex_unlock(&thread_lock);
+    return halide_error_code_success;
 }
 
 }  // extern "C"
@@ -394,20 +389,24 @@ namespace Metal {
 class MetalContextHolder {
     objc_id pool;
     void *const user_context;
+    int status;  // must always be a valid halide_error_code_t value
 
 public:
     mtl_device *device;
     mtl_command_queue *queue;
-    int error;
 
     ALWAYS_INLINE MetalContextHolder(void *user_context, bool create)
         : pool(create_autorelease_pool()), user_context(user_context) {
-        error = halide_metal_acquire_context(user_context, &device, &queue, create);
+        status = halide_metal_acquire_context(user_context, &device, &queue, create);
     }
     ALWAYS_INLINE ~MetalContextHolder() {
-        halide_metal_release_context(user_context);
+        (void)halide_metal_release_context(user_context);  // ignore errors
         drain_autorelease_pool(pool);
     }
+
+    ALWAYS_INLINE int error() const {
+        return status;
+    }
 };
 
 struct command_buffer_completed_handler_block_descriptor_1 {
@@ -456,28 +455,28 @@ WEAK int halide_metal_device_malloc(void *user_context, halide_buffer_t *buf) {
 
     size_t size = buf->size_in_bytes();
     if (size == 0) {
-        error(user_context) << "halide_metal_device_malloc: Failed to allocate buffer of size 0.\n";
-        return -1;
+        error(user_context) << "halide_metal_device_malloc: Failed to allocate buffer of size 0.";
+        return halide_error_code_generic_error;
     }
 
     if (buf->device) {
         // This buffer already has a device allocation
-        return 0;
+        return halide_error_code_success;
     }
 
     // Check all strides positive
     for (int i = 0; i < buf->dimensions; i++) {
         if (buf->dim[i].stride < 0) {
-            error(user_context) << "halide_metal_device_malloc: negatives strides are illegal.\n";
-            return -1;
+            error(user_context) << "halide_metal_device_malloc: negatives strides are illegal.";
+            return halide_error_code_generic_error;
         }
     }
 
     debug(user_context) << "    allocating " << *buf << "\n";
 
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
 
 #ifdef DEBUG_RUNTIME
@@ -493,7 +492,7 @@ WEAK int halide_metal_device_malloc(void *user_context, halide_buffer_t *buf) {
     if (metal_buf == nullptr) {
         free(handle);
         error(user_context) << "Metal: Failed to allocate buffer of size " << (int64_t)size << ".\n";
-        return -1;
+        return halide_error_code_out_of_memory;
     }
 
     handle->buf = metal_buf;
@@ -508,14 +507,14 @@ WEAK int halide_metal_device_malloc(void *user_context, halide_buffer_t *buf) {
     debug(user_context) << "    Time: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_device_free(void *user_context, halide_buffer_t *buf) {
     debug(user_context) << "halide_metal_device_free called on buf "
                         << buf << " device is " << buf->device << "\n";
     if (buf->device == 0) {
-        return 0;
+        return halide_error_code_success;
     }
 
 #ifdef DEBUG_RUNTIME
@@ -524,8 +523,8 @@ WEAK int halide_metal_device_free(void *user_context, halide_buffer_t *buf) {
 
     device_handle *handle = (device_handle *)buf->device;
     if (((device_handle *)buf->device)->offset != 0) {
-        error(user_context) << "halide_metal_device_free: halide_metal_device_free called on buffer obtained from halide_device_crop.\n";
-        return -1;
+        error(user_context) << "halide_metal_device_free: halide_metal_device_free called on buffer obtained from halide_device_crop.";
+        return halide_error_code_generic_error;
     }
 
     release_ns_object(handle->buf);
@@ -539,15 +538,14 @@ WEAK int halide_metal_device_free(void *user_context, halide_buffer_t *buf) {
     debug(user_context) << "    Time: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_initialize_kernels(void *user_context, void **state_ptr, const char *source, int source_size) {
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
-
 #ifdef DEBUG_RUNTIME
     uint64_t t_before = halide_current_time_ns(user_context);
 #endif
@@ -566,12 +564,12 @@ WEAK int halide_metal_initialize_kernels(void *user_context, void **state_ptr, c
     debug(user_context) << "Time for halide_metal_initialize_kernels: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK void halide_metal_finalize_kernels(void *user_context, void *state_ptr) {
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error == 0) {
+    if (metal_context.error() == halide_error_code_success) {
         compilation_cache.release_hold(user_context, metal_context.device, state_ptr);
     }
 }
@@ -601,8 +599,8 @@ WEAK int halide_metal_device_sync(void *user_context, struct halide_buffer_t *bu
 #endif
 
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
 
     halide_metal_device_sync_internal(metal_context.queue, buffer);
@@ -612,18 +610,17 @@ WEAK int halide_metal_device_sync(void *user_context, struct halide_buffer_t *bu
     debug(user_context) << "Time for halide_metal_device_sync: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_device_release(void *user_context) {
     // The MetalContext object does not allow the context storage to be modified,
     // so we use halide_metal_acquire_context directly.
-    int error;
     mtl_device *acquired_device;
     mtl_command_queue *acquired_queue;
-    error = halide_metal_acquire_context(user_context, &acquired_device, &acquired_queue, false);
-    if (error != 0) {
-        return error;
+    auto result = halide_metal_acquire_context(user_context, &acquired_device, &acquired_queue, false);
+    if (result) {
+        return result;
     }
 
     if (acquired_device) {
@@ -644,9 +641,7 @@ WEAK int halide_metal_device_release(void *user_context) {
         }
     }
 
-    halide_metal_release_context(user_context);
-
-    return 0;
+    return halide_metal_release_context(user_context);
 }
 
 WEAK int halide_metal_copy_to_device(void *user_context, halide_buffer_t *buffer) {
@@ -655,13 +650,13 @@ WEAK int halide_metal_copy_to_device(void *user_context, halide_buffer_t *buffer
 #endif
 
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
 
     if (!(buffer->host && buffer->device)) {
-        error(user_context) << "halide_metal_copy_to_device: either host or device is null.\n";
-        return -1;
+        error(user_context) << "halide_metal_copy_to_device: either host or device is null.";
+        return halide_error_code_generic_error;
     }
 
     device_copy c = make_host_to_device_copy(buffer);
@@ -690,7 +685,7 @@ WEAK int halide_metal_copy_to_device(void *user_context, halide_buffer_t *buffer
                         << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_copy_to_host(void *user_context, halide_buffer_t *buffer) {
@@ -699,20 +694,20 @@ WEAK int halide_metal_copy_to_host(void *user_context, halide_buffer_t *buffer)
 #endif
 
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
 
     halide_metal_device_sync_internal(metal_context.queue, buffer);
 
     if (!(buffer->host && buffer->device)) {
-        error(user_context) << "halide_metal_copy_to_host: either host or device is null.\n";
-        return -1;
+        error(user_context) << "halide_metal_copy_to_host: either host or device is null.";
+        return halide_error_code_generic_error;
     }
 
     if (buffer->dimensions > MAX_COPY_DIMS) {
-        error(user_context) << "halide_metal_copy_to_host: buffer->dimensions > MAX_COPY_DIMS.\n";
-        return -1;
+        error(user_context) << "halide_metal_copy_to_host: buffer->dimensions > MAX_COPY_DIMS.";
+        return halide_error_code_generic_error;
     }
 
     device_copy c = make_device_to_host_copy(buffer);
@@ -725,7 +720,7 @@ WEAK int halide_metal_copy_to_host(void *user_context, halide_buffer_t *buffer)
     debug(user_context) << "Time for halide_metal_copy_to_host: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_run(void *user_context,
@@ -734,7 +729,7 @@ WEAK int halide_metal_run(void *user_context,
                           int blocksX, int blocksY, int blocksZ,
                           int threadsX, int threadsY, int threadsZ,
                           int shared_mem_bytes,
-                          size_t arg_sizes[],
+                          halide_type_t arg_types[],
                           void *args[],
                           int8_t arg_is_buffer[]) {
 #ifdef DEBUG_RUNTIME
@@ -742,55 +737,64 @@ WEAK int halide_metal_run(void *user_context,
 #endif
 
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
 
     mtl_command_buffer *command_buffer = new_command_buffer(metal_context.queue, entry_name, strlen(entry_name));
     if (command_buffer == nullptr) {
-        error(user_context) << "Metal: Could not allocate command buffer.\n";
-        return -1;
+        error(user_context) << "Metal: Could not allocate command buffer.";
+        return halide_error_code_generic_error;
     }
 
     mtl_compute_command_encoder *encoder = new_compute_command_encoder(command_buffer);
     if (encoder == nullptr) {
-        error(user_context) << "Metal: Could not allocate compute command encoder.\n";
-        return -1;
+        error(user_context) << "Metal: Could not allocate compute command encoder.";
+        return halide_error_code_generic_error;
     }
 
     mtl_library *library{};
     bool found = compilation_cache.lookup(metal_context.device, state_ptr, library);
     if (!(found && library != nullptr)) {
-        error(user_context) << "Metal: cache lookup failed to find library.\n";
-        return -1;
+        error(user_context) << "Metal: cache lookup failed to find library.";
+        return halide_error_code_generic_error;
     }
 
     mtl_function *function = new_function_with_name(library, entry_name, strlen(entry_name));
     if (function == nullptr) {
-        error(user_context) << "Metal: Could not get function " << entry_name << "from Metal library.\n";
-        return -1;
+        error(user_context) << "Metal: Could not get function " << entry_name << "from Metal library.";
+        return halide_error_code_generic_error;
     }
 
     mtl_compute_pipeline_state *pipeline_state = new_compute_pipeline_state_with_function(metal_context.device, function);
     if (pipeline_state == nullptr) {
-        error(user_context) << "Metal: Could not allocate pipeline state.\n";
-        return -1;
+        error(user_context) << "Metal: Could not allocate pipeline state.";
+        return halide_error_code_generic_error;
     }
 
 #ifdef DEBUG_RUNTIME
     int64_t max_total_threads_per_threadgroup = get_max_total_threads_per_threadgroup(pipeline_state);
     if (max_total_threads_per_threadgroup < threadsX * threadsY * threadsZ) {
-        error(user_context) << "Metal: threadsX(" << threadsX << ") * threadsY(" << threadsY << ") * threadsZ(" << threadsZ << ") (" << (threadsX * threadsY * threadsZ) << ") must be <= " << max_total_threads_per_threadgroup << ". (device threadgroup size limit)\n";
         end_encoding(encoder);
         release_ns_object(pipeline_state);
-        return -1;
+        error(user_context) << "Metal: threadsX(" << threadsX << ") * threadsY("
+                            << threadsY << ") * threadsZ(" << threadsZ
+                            << ") (" << (threadsX * threadsY * threadsZ)
+                            << ") must be <= " << max_total_threads_per_threadgroup
+                            << ". (device threadgroup size limit)\n";
+        return halide_error_code_generic_error;
     }
 #endif
 
     set_compute_pipeline_state(encoder, pipeline_state);
 
+    int num_kernel_args = 0;
+    for (int i = 0; arg_types[i].as_u32() != 0; i++) {
+        ++num_kernel_args;
+    }
+
     size_t total_args_size = 0;
-    for (size_t i = 0; arg_sizes[i] != 0; i++) {
+    for (int i = 0; i < num_kernel_args; i++) {
         if (!arg_is_buffer[i]) {
             // Metal requires natural alignment for all types in structures.
             // Assert arg_size is exactly a power of two and adjust size to start
@@ -799,9 +803,10 @@ WEAK int halide_metal_run(void *user_context,
             // TODO(zalman): This seems fishy - if the arguments are
             // not already sorted in decreasing order of size, wrong
             // results occur. To repro, remove the sorting code in CodeGen_GPU_Host
-            halide_debug_assert(user_context, (arg_sizes[i] & (arg_sizes[i] - 1)) == 0);
-            total_args_size = (total_args_size + arg_sizes[i] - 1) & ~(arg_sizes[i] - 1);
-            total_args_size += arg_sizes[i];
+            size_t arg_size = arg_types[i].bytes();
+            halide_debug_assert(user_context, (arg_size & (arg_size - 1)) == 0);
+            total_args_size = (total_args_size + arg_size - 1) & ~(arg_size - 1);
+            total_args_size += arg_size;
         }
     }
 
@@ -830,18 +835,19 @@ WEAK int halide_metal_run(void *user_context,
         } else {
             args_buffer = new_buffer(metal_context.device, padded_args_size);
             if (args_buffer == nullptr) {
-                error(user_context) << "Metal: Could not allocate arguments buffer.\n";
                 release_ns_object(pipeline_state);
-                return -1;
+                error(user_context) << "Metal: Could not allocate arguments buffer.";
+                return halide_error_code_generic_error;
             }
             args_ptr = (char *)buffer_contents(args_buffer);
         }
         size_t offset = 0;
-        for (size_t i = 0; arg_sizes[i] != 0; i++) {
+        for (int i = 0; i < num_kernel_args; i++) {
             if (!arg_is_buffer[i]) {
-                memcpy(&args_ptr[offset], args[i], arg_sizes[i]);
-                offset = (offset + arg_sizes[i] - 1) & ~(arg_sizes[i] - 1);
-                offset += arg_sizes[i];
+                size_t arg_size = arg_types[i].bytes();
+                memcpy(&args_ptr[offset], args[i], arg_size);
+                offset = (offset + arg_size - 1) & ~(arg_size - 1);
+                offset += arg_size;
             }
         }
         halide_debug_assert(user_context, offset == total_args_size);
@@ -855,9 +861,8 @@ WEAK int halide_metal_run(void *user_context,
         buffer_index++;
     }
 
-    for (size_t i = 0; arg_sizes[i] != 0; i++) {
+    for (int i = 0; i < num_kernel_args; i++) {
         if (arg_is_buffer[i]) {
-            halide_debug_assert(user_context, arg_sizes[i] == sizeof(uint64_t));
             device_handle *handle = (device_handle *)((halide_buffer_t *)args[i])->device;
             set_input_buffer(encoder, handle->buf, handle->offset, buffer_index);
             buffer_index++;
@@ -892,41 +897,43 @@ WEAK int halide_metal_run(void *user_context,
     debug(user_context) << "Time for halide_metal_device_run: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_device_and_host_malloc(void *user_context, struct halide_buffer_t *buffer) {
     debug(user_context) << "halide_metal_device_and_host_malloc called.\n";
-    int result = halide_metal_device_malloc(user_context, buffer);
-    if (result == 0) {
-        mtl_buffer *metal_buffer = ((device_handle *)(buffer->device))->buf;
-        buffer->host = (uint8_t *)buffer_contents(metal_buffer);
-        debug(user_context) << "halide_metal_device_and_host_malloc"
-                            << " device = " << (void *)buffer->device
-                            << " metal_buffer = " << metal_buffer
-                            << " host = " << buffer->host << "\n";
+    auto result = halide_metal_device_malloc(user_context, buffer);
+    if (result) {
+        return result;
     }
-    return result;
+
+    mtl_buffer *metal_buffer = ((device_handle *)(buffer->device))->buf;
+    buffer->host = (uint8_t *)buffer_contents(metal_buffer);
+    debug(user_context) << "halide_metal_device_and_host_malloc"
+                        << " device = " << (void *)buffer->device
+                        << " metal_buffer = " << metal_buffer
+                        << " host = " << buffer->host << "\n";
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_device_and_host_free(void *user_context, struct halide_buffer_t *buffer) {
     debug(user_context) << "halide_metal_device_and_host_free called.\n";
-    halide_metal_device_free(user_context, buffer);
+    auto result = halide_metal_device_free(user_context, buffer);
     buffer->host = nullptr;
-    return 0;
+    return result;
 }
 
 WEAK int halide_metal_buffer_copy(void *user_context, struct halide_buffer_t *src,
                                   const struct halide_device_interface_t *dst_device_interface,
                                   struct halide_buffer_t *dst) {
     if (dst->dimensions > MAX_COPY_DIMS) {
-        error(user_context) << "Buffer has too many dimensions to copy to/from GPU\n";
+        error(user_context) << "Buffer has too many dimensions to copy to/from GPU";
         return halide_error_code_device_buffer_copy_failed;
     }
 
     // We only handle copies to metal buffers or to host
     if (dst_device_interface != nullptr && dst_device_interface != &metal_device_interface) {
-        error(user_context) << "halide_metal_buffer_copy: only handle copies to metal buffers or to host\n";
+        error(user_context) << "halide_metal_buffer_copy: only handle copies to metal buffers or to host";
         return halide_error_code_device_buffer_copy_failed;
     }
 
@@ -942,21 +949,20 @@ WEAK int halide_metal_buffer_copy(void *user_context, struct halide_buffer_t *sr
     bool to_host = !dst_device_interface;
 
     if (!(from_host || src->device)) {
-        error(user_context) << "halide_metal_buffer_copy: invalid copy source\n";
+        error(user_context) << "halide_metal_buffer_copy: invalid copy source";
         return halide_error_code_device_buffer_copy_failed;
     }
     if (!(to_host || dst->device)) {
-        error(user_context) << "halide_metal_buffer_copy: invalid copy destination\n";
+        error(user_context) << "halide_metal_buffer_copy: invalid copy destination";
         return halide_error_code_device_buffer_copy_failed;
     }
 
     device_copy c = make_buffer_copy(src, from_host, dst, to_host);
 
-    int err = 0;
     {
         MetalContextHolder metal_context(user_context, true);
-        if (metal_context.error != 0) {
-            return metal_context.error;
+        if (metal_context.error()) {
+            return metal_context.error();
         }
 
         debug(user_context)
@@ -1019,7 +1025,7 @@ WEAK int halide_metal_buffer_copy(void *user_context, struct halide_buffer_t *sr
 #endif
     }
 
-    return err;
+    return halide_error_code_success;
 }
 
 namespace {
@@ -1029,14 +1035,14 @@ WEAK int metal_device_crop_from_offset(void *user_context,
                                        int64_t offset,
                                        struct halide_buffer_t *dst) {
     MetalContextHolder metal_context(user_context, true);
-    if (metal_context.error != 0) {
-        return metal_context.error;
+    if (metal_context.error()) {
+        return metal_context.error();
     }
 
     dst->device_interface = src->device_interface;
     device_handle *new_handle = (device_handle *)malloc(sizeof(device_handle));
     if (new_handle == nullptr) {
-        error(user_context) << "halide_metal_device_crop: malloc failed making device handle.\n";
+        error(user_context) << "halide_metal_device_crop: malloc failed making device handle.";
         return halide_error_code_out_of_memory;
     }
 
@@ -1044,7 +1050,7 @@ WEAK int metal_device_crop_from_offset(void *user_context,
     new_handle->buf = ((device_handle *)src->device)->buf;
     new_handle->offset = ((device_handle *)src->device)->offset + offset;
     dst->device = (uint64_t)new_handle;
-    return 0;
+    return halide_error_code_success;
 }
 
 }  // namespace
@@ -1072,7 +1078,7 @@ WEAK int halide_metal_device_release_crop(void *user_context,
     debug(user_context) << "halide_metal_device_release_crop called on buf "
                         << buf << " device is " << buf->device << "\n";
     if (buf->device == 0) {
-        return 0;
+        return halide_error_code_success;
     }
 
 #ifdef DEBUG_RUNTIME
@@ -1089,17 +1095,17 @@ WEAK int halide_metal_device_release_crop(void *user_context,
     debug(user_context) << "    Time: " << (t_after - t_before) / 1.0e6 << " ms\n";
 #endif
 
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_wrap_buffer(void *user_context, struct halide_buffer_t *buf, uint64_t buffer) {
     if (buf->device != 0) {
-        error(user_context) << "halide_metal_wrap_buffer: device is nonzero.\n";
-        return -2;
+        error(user_context) << "halide_metal_wrap_buffer: device field is already non-zero.";
+        return halide_error_code_generic_error;
     }
     device_handle *handle = (device_handle *)malloc(sizeof(device_handle));
     if (handle == nullptr) {
-        error(user_context) << "halide_metal_wrap_buffer: malloc failed making device handle.\n";
+        error(user_context) << "halide_metal_wrap_buffer: malloc failed making device handle.";
         return halide_error_code_out_of_memory;
     }
     handle->buf = (mtl_buffer *)buffer;
@@ -1108,22 +1114,22 @@ WEAK int halide_metal_wrap_buffer(void *user_context, struct halide_buffer_t *bu
     buf->device = (uint64_t)handle;
     buf->device_interface = &metal_device_interface;
     buf->device_interface->impl->use_module();
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK int halide_metal_detach_buffer(void *user_context, struct halide_buffer_t *buf) {
     if (buf->device == 0) {
-        return 0;
+        return halide_error_code_success;
     }
     if (buf->device_interface != &metal_device_interface) {
-        error(user_context) << "halide_metal_detach_buffer: device is not metal.\n";
-        return -2;
+        error(user_context) << "halide_metal_detach_buffer: device is not metal.";
+        return halide_error_code_generic_error;
     }
     buf->device_interface->impl->release_module();
     buf->device_interface = nullptr;
     free((device_handle *)buf->device);
     buf->device = 0;
-    return 0;
+    return halide_error_code_success;
 }
 
 WEAK uintptr_t halide_metal_get_buffer(void *user_context, struct halide_buffer_t *buf) {
@@ -1149,7 +1155,7 @@ WEAK const struct halide_device_interface_t *halide_metal_device_interface() {
 namespace {
 WEAK __attribute__((destructor)) void halide_metal_cleanup() {
     compilation_cache.release_all(nullptr, release_ns_object);
-    halide_metal_device_release(nullptr);
+    (void)halide_metal_device_release(nullptr);  // ignore errors
 }
 }  // namespace
 
