@@ -9,6 +9,8 @@
 #include "HalideRuntimeOpenCL.h"
 #include "HalideRuntimeOpenGLCompute.h"
 #include "HalideRuntimeQurt.h"
+#include "HalideRuntimeVulkan.h"
+#include "HalideRuntimeWebGPU.h"
 #include "cpu_features.h"
 
 // This runtime module will contain extern declarations of the Halide
@@ -18,6 +20,8 @@
 // Can be generated via the following:
 // cat src/runtime/runtime_internal.h src/runtime/HalideRuntime*.h | grep "^[^ ][^(]*halide_[^ ]*(" | grep -v '#define' | sed "s/[^(]*halide/halide/" | sed "s/(.*//" | sed "s/^h/    \(void *)\&h/" | sed "s/$/,/" | sort | uniq
 
+extern "C" void halide_unused_force_include_types();
+
 extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_buffer_copy,
     (void *)&halide_buffer_to_string,
@@ -97,6 +101,7 @@ extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_hexagon_device_release,
     (void *)&halide_hexagon_get_device_handle,
     (void *)&halide_hexagon_get_device_size,
+    (void *)&halide_hexagon_get_module_state,
     (void *)&halide_hexagon_initialize_kernels,
     (void *)&halide_hexagon_finalize_kernels,
     (void *)&halide_hexagon_power_hvx_off,
@@ -111,7 +116,6 @@ extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_join_thread,
     (void *)&halide_load_library,
     (void *)&halide_malloc,
-    (void *)&halide_matlab_call_pipeline,
     (void *)&halide_memoization_cache_cleanup,
     (void *)&halide_memoization_cache_evict,
     (void *)&halide_memoization_cache_lookup,
@@ -210,4 +214,14 @@ extern "C" __attribute__((used)) void *halide_runtime_api_functions[] = {
     (void *)&halide_d3d12compute_finalize_kernels,
     (void *)&halide_d3d12compute_release_context,
     (void *)&halide_d3d12compute_run,
+    (void *)&halide_vulkan_acquire_context,
+    (void *)&halide_vulkan_device_interface,
+    (void *)&halide_vulkan_initialize_kernels,
+    (void *)&halide_vulkan_release_context,
+    (void *)&halide_vulkan_run,
+    (void *)&halide_webgpu_device_interface,
+    (void *)&halide_webgpu_initialize_kernels,
+    (void *)&halide_webgpu_finalize_kernels,
+    (void *)&halide_webgpu_run,
+    (void *)&halide_unused_force_include_types,
 };
