@@ -6,17 +6,11 @@
 
 using namespace Halide;
 
-#ifdef _WIN32
-#define DLLEXPORT __declspec(dllexport)
-#else
-#define DLLEXPORT
-#endif
-
 // External functions to track whether the cache is working.
 
 int call_count = 0;
 
-extern "C" DLLEXPORT int count_calls(halide_buffer_t *out) {
+extern "C" HALIDE_EXPORT_SYMBOL int count_calls(halide_buffer_t *out) {
     if (!out->is_bounds_query()) {
         call_count++;
         Halide::Runtime::Buffer<uint8_t>(*out).fill(42);
@@ -26,7 +20,7 @@ extern "C" DLLEXPORT int count_calls(halide_buffer_t *out) {
 
 int call_count_with_arg = 0;
 
-extern "C" DLLEXPORT int count_calls_with_arg(uint8_t val, halide_buffer_t *out) {
+extern "C" HALIDE_EXPORT_SYMBOL int count_calls_with_arg(uint8_t val, halide_buffer_t *out) {
     if (!out->is_bounds_query()) {
         call_count_with_arg++;
         Halide::Runtime::Buffer<uint8_t>(*out).fill(val);
@@ -36,7 +30,7 @@ extern "C" DLLEXPORT int count_calls_with_arg(uint8_t val, halide_buffer_t *out)
 
 int call_count_with_arg_parallel[8];
 
-extern "C" DLLEXPORT int count_calls_with_arg_parallel(uint8_t val, halide_buffer_t *out) {
+extern "C" HALIDE_EXPORT_SYMBOL int count_calls_with_arg_parallel(uint8_t val, halide_buffer_t *out) {
     if (!out->is_bounds_query()) {
         call_count_with_arg_parallel[out->dim[2].min]++;
         Halide::Runtime::Buffer<uint8_t>(*out).fill(val);
@@ -46,7 +40,7 @@ extern "C" DLLEXPORT int count_calls_with_arg_parallel(uint8_t val, halide_buffe
 
 int call_count_staged[4];
 
-extern "C" DLLEXPORT int count_calls_staged(int32_t stage, uint8_t val, halide_buffer_t *in, halide_buffer_t *out) {
+extern "C" HALIDE_EXPORT_SYMBOL int count_calls_staged(int32_t stage, uint8_t val, halide_buffer_t *in, halide_buffer_t *out) {
     if (in->is_bounds_query()) {
         for (int i = 0; i < out->dimensions; i++) {
             in->dim[i] = out->dim[i];
@@ -60,7 +54,7 @@ extern "C" DLLEXPORT int count_calls_staged(int32_t stage, uint8_t val, halide_b
     return 0;
 }
 
-extern "C" DLLEXPORT int computed_eviction_key(int a) {
+extern "C" HALIDE_EXPORT_SYMBOL int computed_eviction_key(int a) {
     return 2020 + a;
 }
 HalideExtern_1(int, computed_eviction_key, int);
