@@ -2,7 +2,7 @@
 
 // So far we have written Halide schedules by hand, but it is also possible to
 // ask Halide to suggest a reasonable schedule. We call this auto-scheduling.
-// This lesson demonstrates how to use the auto-scheduler to generate a
+// This lesson demonstrates how to use the autoscheduler to generate a
 // copy-pasteable CPU schedule that can be subsequently improved upon.
 
 // On linux or os x, you can compile and run it like so:
@@ -11,7 +11,7 @@
 // export LD_LIBRARY_PATH=<path/to/libHalide.so>   # For linux
 // export DYLD_LIBRARY_PATH=<path/to/libHalide.dylib> # For OS X
 // ./lesson_21_generate -o . -g auto_schedule_gen -f auto_schedule_false -e static_library,h,schedule target=host auto_schedule=false
-// ./lesson_21_generate -o . -g auto_schedule_gen -f auto_schedule_true -e static_library,h,schedule -p <path/to/libautoschedule_mullapudi2016.so> -S Mullapudi2016 target=host auto_schedule=true machine_params=32,16777216,40
+// ./lesson_21_generate -o . -g auto_schedule_gen -f auto_schedule_true -e static_library,h,schedule -p <path/to/libautoschedule_mullapudi2016.so> -S Mullapudi2016 target=host autoscheduler=Mullapudi2016 autoscheduler.parallelism=32 autoscheduler.last_level_cache_size=16777216 autoscheduler.balance=40
 // g++ lesson_21_auto_scheduler_run.cpp -std=c++17 -I <path/to/Halide.h> -I <path/to/tools/halide_image_io.h> auto_schedule_false.a auto_schedule_true.a -ldl -lpthread -o lesson_21_run
 // ./lesson_21_run
 
@@ -64,13 +64,13 @@ public:
         det(x, y) = Sxx(x, y) * Syy(x, y) - Sxy(x, y) * Sxy(x, y);
         trace(x, y) = Sxx(x, y) + Syy(x, y);
         harris(x, y) = det(x, y) - 0.04f * trace(x, y) * trace(x, y);
-        output1(x, y) = harris(x + 2, y + 2);
-        output2(x, y) = factor * harris(x + 2, y + 2);
+        output1(x, y) = harris(x, y);
+        output2(x, y) = factor * harris(x, y);
     }
 
     void schedule() {
-        if (auto_schedule) {
-            // The auto-scheduler requires estimates on all the input/output
+        if (using_autoscheduler()) {
+            // The autoscheduler requires estimates on all the input/output
             // sizes and parameter values in order to compare different
             // alternatives and decide on a good schedule.
 
@@ -95,31 +95,33 @@ public:
             // schedule will be.
 
             // To auto-schedule the pipeline, we don't have to do anything else:
-            // every Generator implicitly has a GeneratorParam named "auto_schedule";
-            // if this is set to true, Halide will call auto_schedule() on all of
-            // our pipeline's outputs automatically.
+            // every Generator implicitly has a GeneratorParam named "auto_scheduler.name";
+            // if this is set to the name of the Autoscheduler we want to use, Halide will
+            // apply it to all of our pipeline's outputs automatically.
 
-            // Every Generator also implicitly has a GeneratorParams named "machine_params",
-            // which allows you to specify characteristics of the machine architecture
-            // for the auto-scheduler; it's generally specified in your Makefile.
+            // Every Generator also implicitly has additional, optional GeneratorParams that are
+            // dependent on the specific Autoscheduler select, which allows you to specify
+            // characteristics of the machine architecture
+            // for the autoscheduler; it's generally specified in your Makefile.
             // If none is specified, the default machine parameters for a generic CPU
-            // architecture will be used by the auto-scheduler.
+            // architecture will be used by the autoscheduler.
 
-            // Let's see some arbitrary but plausible values for the machine parameters.
+            // Let's see some arbitrary but plausible values for the machine parameters
+            // for the Mullapudi2016 Autoscheduler:
             //
-            //      const int kParallelism = 32;
-            //      const int kLastLevelCacheSize = 16 * 1024 * 1024;
-            //      const int kBalance = 40;
-            //      MachineParams machine_params(kParallelism, kLastLevelCacheSize, kBalance);
+            //      autoscheduler=Mullapudi2016
+            //      autoscheduler.parallelism=32
+            //      autoscheduler.last_level_cache_size=16777216
+            //      autoscheduler.balance=40
             //
-            // The arguments to MachineParams are the maximum level of parallelism
-            // available, the size of the last-level cache (in KB), and the ratio
+            // These are the maximum level of parallelism
+            // available, the size of the last-level cache (in bytes), and the ratio
             // between the cost of a miss at the last level cache and the cost
             // of arithmetic on the target architecture, in that order.
 
-            // Note that when using the auto-scheduler, no schedule should have
-            // been applied to the pipeline; otherwise, the auto-scheduler will
-            // throw an error. The current auto-scheduler cannot handle a
+            // Note that when using the autoscheduler, no schedule should have
+            // been applied to the pipeline; otherwise, the autoscheduler will
+            // throw an error. The current autoscheduler cannot handle a
             // partially-scheduled pipeline.
 
             // If HL_DEBUG_CODEGEN is set to 3 or greater, the schedule will be dumped
@@ -131,12 +133,12 @@ public:
             // Halide C++ source, which is readily copy-pasteable back into
             // this very same source file with few modifications. Programmers
             // can use this as a starting schedule and iteratively improve the
-            // schedule. Note that the current auto-scheduler is only able to
+            // schedule. Note that the current autoscheduler is only able to
             // generate CPU schedules and only does tiling, simple vectorization
             // and parallelization. It doesn't deal with line buffering, storage
             // reordering, or factoring reductions.
 
-            // At the time of writing, the auto-scheduler will produce the
+            // At the time of writing, the autoscheduler will produce the
             // following schedule for the estimates and machine parameters
             // declared above when run on this pipeline:
             //
@@ -211,7 +213,7 @@ public:
 
         } else {
             // This is where you would declare the schedule you have written by
-            // hand or paste the schedule generated by the auto-scheduler.
+            // hand or paste the schedule generated by the autoscheduler.
             // We will use a naive schedule here to compare the performance of
             // the autoschedule with a basic schedule.
             gray.compute_root();
